"""
@file
@brief OnnxInferenceNode definition.
"""
import sys
import pprint
import numpy
from onnx import onnx_pb as onnx_proto
from .ops import load_op


class OnnxInferenceNode:
    """
    A node to execute.
    """

    def __init__(self, onnx_node, desc, global_index):
        """
        @param      onnx_node       onnx_node
        @param      desc            internal description
        @param      global_index    it is a function which returns a unique index
                                    for the output this operator generates
        """
        if desc is None:
            raise ValueError("desc should not be None.")  # pragma: no cover
        self.desc = desc
        self.onnx_node = onnx_node
        self._init(global_index)

    @property
    def name(self):
        "Returns the ONNX name."
        return "_".join(
            [self.desc['domain'], self.onnx_node.op_type]).replace(
                ".", "_").replace('__', '_').strip('_')

    def _init(self, global_index):
        """
        Prepares the node.
        """
        self.op_type = self.onnx_node.op_type
        self.order = -1
        self.variable_to_clean = []
        self.inputs = list(self.onnx_node.input)
        self.outputs = list(self.onnx_node.output)
        self.inplaces = []
        self.inputs_indices = [global_index(name) for name in self.inputs]
        self.outputs_indices = [global_index(name) for name in self.outputs]

    def set_order(self, order):
        """
        Defines the order of execution.
        """
        self.order = order

    def add_variable_to_clean(self, name):
        """
        Adds a variable which can be cleaned after the node
        execution.
        """
        self.variable_to_clean.append(name)

    def __str__(self):
        "usual"
        return "Onnx-{}({}) -> {}".format(
            self.op_type, ", ".join(self.inputs),
            ", ".join(self.outputs))

    def __repr__(self):
        "usual"
        return self.__str__()

    def setup_runtime(self, runtime=None, variables=None, rt_class=None,
                      target_opset=None, dtype=None, domain=None,
                      ir_version=None, runtime_options=None):
        """
        Loads runtime.

        @param      runtime         runtime options
        @param      variables       registered variables created by previous operators
        @param      rt_class        runtime class used to compute
                                    prediction of subgraphs
        @param      target_opset    use a specific target opset
        @param      dtype           float computational type
        @param      domain          node domain
        @param      ir_version      if not None, changes the default value
                                    given by :epkg:`ONNX`
        @param      runtime_options runtime options
        """
        if self.desc is None:
            raise AttributeError(
                "desc should not be None.")  # pragma: no cover
        self.preprocess_parameters(
            runtime, rt_class, ir_version=ir_version, target_opset=target_opset)
        options = {'provider': runtime} if runtime else {}
        if domain is not None:
            options['domain'] = domain
        if target_opset is not None:
            options['target_opset'] = target_opset
        if ir_version is not None:
            options['ir_version'] = ir_version
        if runtime_options is not None:
            options.update(runtime_options)
        if runtime == 'onnxruntime2':
            self.ops_ = load_op(self.onnx_node, desc=self.desc,
                                options=options if options else None,
                                variables=variables, dtype=dtype)
        elif runtime in ('python_compiled', 'python_compiled_debug'):
            options['provider'] = 'python'
            self.ops_ = load_op(self.onnx_node, desc=self.desc,
                                options=options if options else None,
                                variables=variables)
        else:
            self.ops_ = load_op(self.onnx_node, desc=self.desc,
                                options=options if options else None,
                                variables=variables)

    def preprocess_parameters(self, runtime, rt_class, ir_version=None,
                              target_opset=None):
        """
        Preprocesses the parameters,
        loads *GraphProto*
        (equivalent to :epkg:`ONNX` graph with
        less metadata).

        @param  runtime         runtime options
        @param  rt_class        runtime class used to compute
                                prediction of subgraphs
        @param  ir_version      if not None, overwrites the default value
        @param  target_opset    use a specific target opset
        """
        if 'atts' not in self.desc:
            return  # pragma: no cover
        for _, v in self.desc['atts'].items():
            if 'value' not in v:
                continue  # pragma: no cover
            value = v['value']
            if isinstance(value, onnx_proto.GraphProto):
                sess = rt_class(v['value'], runtime=runtime,
                                ir_version=ir_version,
                                target_opset=target_opset)
                v['value_rt'] = sess

    def run(self, values):
        """
        Runs the node.
        the function updates values with outputs.

        @param      values      list of existing values
        """
        # This code takes times if the graph contains many nodes.
        # Maybe a C++ container would help in that case (to skip GIL).
        if self.inputs_indices is None:
            args = list(values[k] for k in self.inputs)
        else:
            args = list(values[k] for k in self.inputs_indices)

        try:
            res = self.ops_.run(*args)
        except TypeError as e:
            raise RuntimeError(
                "Unable to run operator %r." % type(self.ops_)) from e

        if not isinstance(res, tuple):
            raise RuntimeError(  # pragma: no cover
                "Results of operator %r should be a tuple." % type(self.ops_))
        if len(self.outputs) != len(res):
            raise RuntimeError(  # pragma: no cover
                "Mismatch number of outputs got {} for names {}.\n{}".format(
                    len(res), list(sorted(self.outputs)),
                    pprint.pformat(self.desc)))

        # This code takes times if the graph contains many nodes.
        # Maybe a C++ container would help in that case (to skip GIL).
        if self.outputs_indices is None:
            for name, value in zip(self.outputs, res):
                values[name] = value
        else:
            for i, r in enumerate(res):
                values[self.outputs_indices[i]] = r

    def switch_initializers_dtype(self, dtype_in=numpy.float32,
                                  dtype_out=numpy.float64):
        """
        Switches all initializers to ``numpy.float64``.
        This only works if the runtime is ``'python'``.

        @param      dtype_in    previous type
        @param      dtype_out   next type
        @return                 done operations
        """
        done = []
        for k, v in self.desc['atts'].items():
            if 'value_rt' not in v:
                continue
            if isinstance(v['value_rt'], numpy.ndarray):
                if v['value_rt'].dtype == dtype_in:
                    v['value_rt'] = v['value_rt'].astype(dtype_out)
                    done.append(("+", "desc", k, v['value_rt']))
                else:
                    done.append(("-", "desc", k, v['value_rt']))
        if hasattr(self, 'ops_') and self.ops_ is not None:
            res = self.ops_.switch_initializers_dtype(dtype_in, dtype_out)
            for r in res:
                done.append(("ops_", ) + r)
        return done

    def _set_shape_inference_runtime(self, values):
        """
        Updates *values* which shapes of the outputs.

        @param      values      container for shapes
        """
        args = [values[k] for k in self.inputs]
        try:
            res = self.ops_.infer_shapes(*args)
        except (TypeError, ValueError) as e:
            raise TypeError(
                "Unable to call infer_shapes with {} arguments for class"
                " '{}' ({})".format(len(args), self.ops_.__class__.__name__,
                                    self.ops_.infer_shapes)) from e
        if not isinstance(res, tuple):
            raise RuntimeError(  # pragma: no cover
                "Results of an operator should be a tuple for operator '{}'"
                ".".format(type(self.ops_)))
        if len(self.outputs) != len(res):
            raise RuntimeError(  # pragma: no cover
                "Mismatch number of outputs got {} != {} for names {} (node='{}')."
                "\n{}".format(
                    len(res), len(self.outputs), list(self.outputs),
                    self.ops_.__class__.__name__,
                    pprint.pformat(self.desc, depth=2)))
        for name, value in zip(self.outputs, res):
            values[name] = value
        return values

    def enable_inplace_compute(self, name):
        """
        Let the node know that one input can be overwritten.

        @param      name        input name
        """
        self.inplaces.append(name)
        self.ops_.enable_inplace_compute(self.inputs.index(name))

    @property
    def inputs_args(self):
        """
        Returns the list of arguments as well as
        the list of parameters with the default values
        (close to the signature).
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        sigs = []
        mand = self.ops_.args_mandatory
        if mand is None:
            mand = self.python_inputs
        sigs.extend(mand)
        if len(self.ops_.args_optional) > 0:
            sigs.extend(self.ops_.args_optional)
            if sys.version_info[:2] >= (3, 8):
                sigs.append('/')
        sigs.extend(self.ops_.args_default)
        return sigs

    @property
    def python_inputs(self):
        """
        Returns the python arguments.
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        if hasattr(self.ops_, 'python_inputs'):
            return self.ops_.python_inputs
        return self.inputs

    @property
    def modified_args(self):
        """
        Returns the list of modified parameters.
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        return self.ops_.args_default_modified

    def to_python(self, inputs):
        """
        Returns a python code for this operator.

        @param      inputs      inputs name
        @return                 imports, python code, both as strings
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        return self.ops_.to_python(inputs)



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_whole*.
"""
from io import BytesIO
import onnx
from onnxruntime import (
    InferenceSession, SessionOptions, RunOptions, GraphOptimizationLevel)
from onnxruntime.capi.onnxruntime_pybind11_state import (  # pylint: disable=E0611
    Fail as OrtFail,
    InvalidGraph as OrtInvalidGraph,
    InvalidArgument as OrtInvalidArgument,
    NotImplemented as OrtNotImplemented,
    RuntimeException as OrtRuntimeException)
from ...tools.asv_options_helper import display_onnx


class OnnxWholeSession:
    """
    Runs the prediction for a single :epkg:`ONNX`,
    it lets the runtime handle the graph logic as well.
    """

    def __init__(self, onnx_data, runtime, runtime_options=None):
        """
        @param      onnx_data       :epkg:`ONNX` model or data
        @param      runtime         runtime to be used,
                                    mostly :epkg:`onnxruntime`
        @param      runtime_options runtime options
        """
        if runtime != 'onnxruntime1':
            raise NotImplementedError(  # pragma: no cover
                "runtime '{}' is not implemented.".format(runtime))
        if hasattr(onnx_data, 'SerializeToString'):
            onnx_data = onnx_data.SerializeToString()
        self.runtime = runtime
        sess_options = SessionOptions()
        self.run_options = RunOptions()
        try:
            sess_options.sessions_log_verbosity_level = 0
        except AttributeError:  # pragma: no cover
            # onnxruntime not recent enough.
            pass
        try:
            self.run_options.run_log_verbosity_level = 0
        except AttributeError:  # pragma: no cover
            # onnxruntime not recent enough.
            pass
        if (runtime_options is not None and
                runtime_options.get('disable_optimisation', False)):
            sess_options.graph_optimization_level = (  # pragma: no cover
                GraphOptimizationLevel.ORT_ENABLE_ALL)
        try:
            self.sess = InferenceSession(onnx_data, sess_options=sess_options)
        except (OrtFail, OrtNotImplemented, OrtInvalidGraph,
                OrtInvalidArgument, OrtRuntimeException, RuntimeError) as e:
            raise RuntimeError(
                "Unable to create InferenceSession due to '{}'\n{}.".format(
                    e, display_onnx(onnx.load(BytesIO(onnx_data))))) from e

    def run(self, inputs):
        """
        Computes the predictions.

        @param      inputs      dictionary *{variable, value}*
        @return                 list of outputs
        """
        return self.sess.run(None, inputs, self.run_options)



"""
@file
@brief Looks into the code and detects error
before finalizing the benchmark.
"""
import ast


class ImperfectPythonCode(RuntimeError):
    """
    Raised if the code shows errors.
    """
    pass


def verify_code(source, exc=True):
    """
    Verifies :epkg:`python` code.

    @param      source      source to look into
    @param      exc         raise an exception or return the list of
                            missing identifiers
    @return                 tuple(missing identifiers, @see cl CodeNodeVisitor)
    """
    node = ast.parse(source)
    v = CodeNodeVisitor()
    v.visit(node)
    assign = v._assign
    imports = v._imports
    names = v._names
    args = v._args
    known = {'super': None}
    for kn in imports:
        known[kn[0]] = kn
    for kn in assign:
        known[kn[0]] = kn
    for kn in args:
        known[kn[0]] = kn
    issues = set()
    for name in names:
        if name[0] not in known:
            issues.add(name[0])
    if exc and len(issues) > 0:
        raise ImperfectPythonCode(
            "Unknown identifiers: {} in\n{}".format(
                ",".join(issues), source))
    return issues, v


class CodeNodeVisitor(ast.NodeVisitor):
    """
    Visits the code, implements verification rules.
    """

    def __init__(self):
        ast.NodeVisitor.__init__(self)
        self._rows = []
        self._indent = 0
        self._stack = []
        self._imports = []
        self._names = []
        self._alias = []
        self._assign = []
        self._args = []
        self._fits = []

    def push(self, row):
        """
        Pushes an element into a list.
        """
        self._rows.append(row)

    def generic_visit(self, node):
        """
        Overrides ``generic_visit`` to check it is not used.
        """
        raise AttributeError(
            "generic_visit_args should be used.")  # pragma: no cover

    def generic_visit_args(self, node, row):
        """
        Overrides ``generic_visit`` to keep track of the indentation
        and the node parent. The function will add field
        ``row["children"] = visited`` nodes from here.

        @param      node        node which needs to be visited
        @param      row         row (a dictionary)
        @return                 See ``ast.NodeVisitor.generic_visit``
        """
        self._indent += 1
        last = len(self._rows)
        res = ast.NodeVisitor.generic_visit(  # pylint: disable=E1111
            self, node)  # pylint: disable=E1111
        row["children"] = [
            _ for _ in self._rows[
                last:] if _["indent"] == self._indent]
        self._indent -= 1
        return res

    def visit(self, node):
        """
        Visits a node, a method must exist for every object class.
        """
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, None)
        if visitor is None:
            if method.startswith("visit_"):
                cont = {
                    "indent": self._indent,
                    "str": method[6:],
                    "node": node}
                self.push(cont)
                return self.generic_visit_args(node, cont)
            raise TypeError("unable to find a method: " +
                            method)  # pragma: no cover
        res = visitor(node)
        # print(method, CodeNodeVisitor.print_node(node))
        return res

    @staticmethod
    def print_node(node):
        """
        Debugging purpose.
        """
        r = []
        for att in ["s", "name", "str", "id", "body", "n",
                    "arg", "targets", "attr", "returns", "ctx"]:
            if att in node.__dict__:
                r.append("{0}={1}".format(att, str(node.__dict__[att])))
        return " ".join(r)

    def print_tree(self):  # pylint: disable=C0116
        """
        Displays the tree of instructions.

        @return     string
        """
        rows = []
        for r in self.Rows:
            rows.append(
                ("{0}{1}: {2}".format(
                    "    " *
                    r["indent"],
                    r["type"],
                    r["str"])))
        return "\n".join(rows)

    @property
    def Rows(self):
        """
        returns a list of dictionaries with all the elements of the code
        """
        return [_ for _ in self._rows if not _.get("remove", False)]

    def visit_Str(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Str",
            "str": node.s,
            "node": node,
            "value": node.s}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Name(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Name",
            "str": node.id,
            "node": node,
            "id": node.id,
            "ctx": node.ctx}
        self.push(cont)
        self._names.append((node.id, node))
        return self.generic_visit_args(node, cont)

    def visit_Expr(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Expr",
            "str": '',
            "node": node,
            "value": node.value}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_alias(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "alias",
            "str": "",
            "node": node,
            "name": node.name,
            "asname": node.asname}
        self.push(cont)
        self._alias.append((node.name, node.asname, node))
        return self.generic_visit_args(node, cont)

    def visit_Module(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Module",
            "str": "",
            "body": node.body,
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Import(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Import",
            "str": "",
            "names": node.names,
            "node": node}
        self.push(cont)
        for name in node.names:
            self._imports.append((name.name, name.asname, node))
        return self.generic_visit_args(node, cont)

    def visit_ImportFrom(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "ImportFrom",
            "str": "",
            "module": node.module,
            "names": node.names,
            "node": node}
        self.push(cont)
        for name in node.names:
            self._imports.append((name.name, name.asname, node.module, node))
        return self.generic_visit_args(node, cont)

    def visit_ClassDef(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "ClassDef",
            "str": "",
            "name": node.name,
            "body": node.body,
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_FunctionDef(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "FunctionDef", "str": node.name, "name": node.name, "body": node.body,
                "node": node, "returns": node.returns}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_arguments(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "arguments", "str": "",
                "node": node, "args": node.args}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_arg(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "arg", "str": node.arg,
                "node": node,
                "arg": node.arg, "annotation": node.annotation}
        self.push(cont)
        self._args.append((node.arg, node))
        return self.generic_visit_args(node, cont)

    def visit_Assign(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "Assign", "str": "", "node": node,
                "targets": node.targets, "value": node.value}
        self.push(cont)
        for t in node.targets:
            if hasattr(t, 'id'):
                self._assign.append((t.id, node))
            else:
                self._assign.append((id(t), node))
        return self.generic_visit_args(node, cont)

    def visit_Store(self, node):  # pylint: disable=C0116
        #cont = { "indent":self._indent, "type": "Store", "str": "" }
        # self.push(cont)
        cont = {}
        return self.generic_visit_args(node, cont)

    def visit_Call(self, node):  # pylint: disable=C0116
        if "attr" in node.func.__dict__:
            cont = {"indent": self._indent, "type": "Call", "str": node.func.attr,
                    "node": node, "func": node.func}
        elif "id" in node.func.__dict__:
            cont = {"indent": self._indent, "type": "Call", "str": node.func.id,
                    "node": node, "func": node.func}
        else:
            cont = {"indent": self._indent, "type": "Call", "str": "",  # pragma: no cover
                    "node": node, "func": node.func}
        self.push(cont)
        if cont['str'] == 'fit':
            self._fits.append(cont)
        return self.generic_visit_args(node, cont)

    def visit_Attribute(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "Attribute", "str": node.attr,
                "node": node, "value": node.value, "ctx": node.ctx, "attr": node.attr}
        self.push(cont)
        # last = len(self._rows)
        res = self.generic_visit_args(node, cont)

        if len(cont["children"]) > 0:
            fir = cont["children"][0]
            if 'type' in fir and fir["type"] == "Name":
                parent = fir["node"].id
                cont["str"] = "{0}.{1}".format(parent, cont["str"])
                cont["children"][0]["remove"] = True
        return res

    def visit_Load(self, node):  # pylint: disable=C0116
        cont = {}
        return self.generic_visit_args(node, cont)

    def visit_keyword(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "keyword", "str": "{0}".format(node.arg),
                "node": node, "arg": node.arg, "value": node.value}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_BinOp(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "BinOp",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_UnaryOp(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "UnaryOp",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Not(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Not",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Invert(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Invert",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_BoolOp(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "BoolOp",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Mult(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Mult",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Div(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Div",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_FloorDiv(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "FloorDiv",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Add(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Add",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Pow(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Pow",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_In(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "In",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_AugAssign(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "AugAssign",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Eq(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Eq",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_IsNot(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "IsNot",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Is(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Is",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_And(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "And",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_BitAnd(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "BitAnd",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Or(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Or",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_NotEq(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "NotEq",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Mod(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Mod",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Sub(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Sub",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_USub(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "USub",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Compare(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Compare",
            "str": "",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Gt(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "Gt", "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_GtE(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "GtE", "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Lt(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "Lt", "str": "", "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Num(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Num",
            "node": node,
            "str": "{0}".format(
                node.n),
            'n': node.n}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Return(self, node):  # pylint: disable=C0116
        cont = {"indent": self._indent, "type": "Return", "node": node, "str": "",
                'value': node.value}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_List(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "List",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_ListComp(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "ListComp",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_comprehension(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "comprehension",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Dict(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Dict",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Tuple(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "Tuple",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_NameConstant(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "type": "NameConstant",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_(self, node):  # pylint: disable=C0116
        raise RuntimeError(  # pragma: no cover
            "This node is not handled: {}".format(node))

    def visit_Subscript(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "Subscript",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_ExtSlice(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "ExtSlice",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Slice(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "Slice",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Index(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "Index",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_If(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "If",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_IfExp(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "IfExp",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_Lambda(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "Lambda",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)

    def visit_GeneratorExp(self, node):  # pylint: disable=C0116
        cont = {
            "indent": self._indent,
            "str": "GeneratorExp",
            "node": node}
        self.push(cont)
        return self.generic_visit_args(node, cont)



"""
@file
@brief Scenario for additional converters.
"""
from lightgbm import LGBMRegressor, LGBMClassifier
from xgboost import XGBRegressor, XGBClassifier


def find_suitable_problem(model):
    """
    Defines suitables problems for additional converters.

    .. runpython::
        :showcode:
        :rst:

        from mlprodict.onnx_conv.validate_scenarios import find_suitable_problem
        from mlprodict.onnxrt.validate.validate_helper import sklearn_operators
        from pyquickhelper.pandashelper import df2rst
        from pandas import DataFrame
        res = sklearn_operators(extended=True)
        res = [_ for _ in res if _['package'] != 'sklearn']
        rows = []
        for model in res:
            name = model['name']
            row = dict(name=name)
            try:
                prob = find_suitable_problem(model['cl'])
                if prob is None:
                    continue
                for p in prob:
                    row[p] = 'X'
            except RuntimeError:
                pass
            rows.append(row)
        df = DataFrame(rows).set_index('name')
        df = df.sort_index()
        print(df2rst(df, index=True))

    """
    def _internal(model):
        # Exceptions
        if model in {LGBMRegressor, XGBRegressor}:
            return ['b-reg', '~b-reg-64']

        if model in {LGBMClassifier, XGBClassifier}:
            return ['b-cl', 'm-cl', '~b-cl-64']

        # Not in this list
        return None

    res = _internal(model)
    return res


def build_custom_scenarios():
    """
    Defines parameters values for some operators.

    .. runpython::
        :showcode:

        from mlprodict.onnx_conv.validate_scenarios import build_custom_scenarios
        import pprint
        pprint.pprint(build_custom_scenarios())
    """
    return {
        # scenarios
        LGBMClassifier: [
            ('default', {'n_estimators': 5}, {'conv_options': [
             {LGBMClassifier: {'zipmap': False}}]}),
        ],
        LGBMRegressor: [
            ('default', {'n_estimators': 100}),
        ],
        XGBClassifier: [
            ('default', {'n_estimators': 5}, {'conv_options': [
             {XGBClassifier: {'zipmap': False}}]}),
        ],
        XGBRegressor: [
            ('default', {'n_estimators': 100}),
        ],
    }



"""
@file
@brief Investigate issues happening with float32.
"""
import numpy
from numpy.random import randint


def astype_range(arr, dtype=numpy.float32, force=1):
    """
    Computes ranges for every number in an array
    once converted into *float32*. The function returns
    two matrices which produces two numbers
    *a* et *b*, the number rounded to float32
    is in interval :math:`[a, b]`.

    @param      arr     array
    @param      dtype   type to convert to
    @param      force   does something like *[i] +/- force |i - [i]|*
    @return             minimum, maximum
    """
    conv = arr.astype(dtype)
    delta = numpy.abs(arr - conv)
    delta = numpy.maximum(numpy.abs(arr) * 1e-7, delta)
    maxa = (conv + delta * force).astype(dtype)
    mina = (conv - delta * force).astype(dtype)
    return mina, maxa


def enumerate_random_inputs(inputs, n=100, dtype=numpy.float32, force=1):
    """
    Enumerates random matrices.

    @param      inputs      inputs (dictionary)
    @param      n           number of iterations
    @param      dtype       type to convert to
    @param      force       does something like *[i] +/- force |i - [i]|*
    """
    keys = list(inputs)
    ranges = {k: astype_range(v, dtype=dtype, force=force)
              for k, v in inputs.items()}
    for _ in range(n):
        new_inputs = {}
        for k in keys:
            rnd = randint(0, 2, inputs[k].size).reshape(  # pylint: disable=E1101
                inputs[k].shape)  # pylint: disable=E1101
            if rnd.min() == rnd.max() or rnd.max() != 1:
                raise RuntimeError(  # pragma: no cover
                    "Minimum and maximum are equal or maximum is not 1. "
                    "Randomness failed.")
            rnd = rnd.astype(dtype)
            ma1 = ranges[k][0] * rnd
            ma2 = ranges[k][1] * (-(rnd - 1))
            inp = (ma1 + ma2)
            new_inputs[k] = inp
        yield new_inputs


def onnx_shaker(oinf, inputs, output_fct, n=100, dtype=numpy.float32, force=1):
    """
    Shakes a model :epkg:`ONNX`.
    Explores the ranges for every prediction.
    Uses @see fn astype_range

    @param      oinf        object of type @see cl OnnxInference
    @param      inputs      inputs
    @param      output_fct  output function which extracts
                            a single array from the output
    @param      dtype       type to convert to
    @param      force       does something like *[i] +/- force |i - [i]|*
    @return                 ranges for each predictions

    See notebook :ref:`onnxshakerrst` for an example of use.
    """
    results = None
    for i, new_inputs in enumerate(enumerate_random_inputs(
            inputs, n=n, dtype=dtype, force=force)):
        res_ = oinf.run(new_inputs)
        res = output_fct(res_)
        sq = numpy.squeeze(res)
        if len(sq.shape) != 1:
            raise ValueError(  # pragma: no cover
                "The function only works with shape={}".format(sq.shape))
        if results is None:
            results = numpy.empty((sq.shape[0], n), dtype=sq.dtype)
        results[:, i] = sq

    results.sort(axis=1)
    return results



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunReduceNumpy


class ReduceLogSumExp(OpRunReduceNumpy):

    atts = {'axes': [], 'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunReduceNumpy.__init__(self, onnx_node, desc=desc,
                                  expected_attributes=ReduceLogSumExp.atts,
                                  **options)

    def _run(self, data):  # pylint: disable=W0221
        tax = tuple(self.axes) if self.axes else None
        data_max = data.copy()
        ind = numpy.isinf(data_max)
        data_max[ind] = -numpy.inf
        mx = data_max.max(axis=tax, keepdims=True)
        sub = numpy.subtract(data, mx)
        exp = numpy.exp(sub, out=sub)
        mxs = numpy.sum(exp, axis=tax,
                        keepdims=True,
                        dtype=data.dtype)
        res = numpy.log(mxs) + mx
        if not self.keepdims:
            res = numpy.squeeze(res, axis=tax)
        return (res, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from collections import OrderedDict
import numpy
from ._op_helper import _get_typed_class_attribute
from ._op import OpRunUnaryNum, RuntimeTypeError
from ._new_ops import OperatorSchema
from .op_tree_ensemble_regressor_ import (  # pylint: disable=E0611,E0401
    RuntimeTreeEnsembleRegressorFloat, RuntimeTreeEnsembleRegressorDouble)
from .op_tree_ensemble_regressor_p_ import (  # pylint: disable=E0611,E0401
    RuntimeTreeEnsembleRegressorPFloat, RuntimeTreeEnsembleRegressorPDouble)


class TreeEnsembleRegressorCommon(OpRunUnaryNum):

    def __init__(self, dtype, onnx_node, desc=None,
                 expected_attributes=None, runtime_version=3, **options):
        OpRunUnaryNum.__init__(
            self, onnx_node, desc=desc,
            expected_attributes=expected_attributes, **options)
        self._init(dtype=dtype, version=runtime_version)

    def _get_typed_attributes(self, k):
        return _get_typed_class_attribute(self, k, self.__class__.atts)

    def _find_custom_operator_schema(self, op_name):
        """
        Finds a custom operator defined by this runtime.
        """
        if op_name == "TreeEnsembleRegressorDouble":
            return TreeEnsembleRegressorDoubleSchema()
        raise RuntimeError(  # pragma: no cover
            "Unable to find a schema for operator '{}'.".format(op_name))

    def _init(self, dtype, version):
        if dtype == numpy.float32:
            if version == 0:
                self.rt_ = RuntimeTreeEnsembleRegressorFloat()
            elif version == 1:
                self.rt_ = RuntimeTreeEnsembleRegressorPFloat(
                    60, 20, False, False)
            elif version == 2:
                self.rt_ = RuntimeTreeEnsembleRegressorPFloat(
                    60, 20, True, False)
            elif version == 3:
                self.rt_ = RuntimeTreeEnsembleRegressorPFloat(
                    60, 20, True, True)
            else:
                raise ValueError("Unknown version '{}'.".format(version))
        elif dtype == numpy.float64:
            if version == 0:
                self.rt_ = RuntimeTreeEnsembleRegressorDouble()
            elif version == 1:
                self.rt_ = RuntimeTreeEnsembleRegressorPDouble(
                    60, 20, False, False)
            elif version == 2:
                self.rt_ = RuntimeTreeEnsembleRegressorPDouble(
                    60, 20, True, False)
            elif version == 3:
                self.rt_ = RuntimeTreeEnsembleRegressorPDouble(
                    60, 20, True, True)
            else:
                raise ValueError("Unknown version '{}'.".format(version))
        else:
            raise RuntimeTypeError(  # pragma: no cover
                "Unsupported dtype={}.".format(dtype))
        atts = [self._get_typed_attributes(k)
                for k in self.__class__.atts]
        self.rt_.init(*atts)

    def _run(self, x):  # pylint: disable=W0221
        """
        This is a C++ implementation coming from
        :epkg:`onnxruntime`.
        `tree_ensemble_classifier.cc
        <https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/ml/tree_ensemble_classifier.cc>`_.
        See class :class:`RuntimeTreeEnsembleRegressorFloat
        <mlprodict.onnxrt.ops_cpu.op_tree_ensemble_regressor_.RuntimeTreeEnsembleRegressorFloat>` or
        class :class:`RuntimeTreeEnsembleRegressorDouble
        <mlprodict.onnxrt.ops_cpu.op_tree_ensemble_regressor_.RuntimeTreeEnsembleRegressorDouble>`.
        """
        if hasattr(x, 'todense'):
            x = x.todense()
        pred = self.rt_.compute(x)
        if pred.shape[0] != x.shape[0]:
            pred = pred.reshape(x.shape[0], pred.shape[0] // x.shape[0])
        return (pred, )


class TreeEnsembleRegressor(TreeEnsembleRegressorCommon):

    atts = OrderedDict([
        ('aggregate_function', b'SUM'),
        ('base_values', numpy.empty(0, dtype=numpy.float32)),
        ('n_targets', 1),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float32)),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float32)),
        ('post_transform', b'NONE'),
        ('target_ids', numpy.empty(0, dtype=numpy.int64)),
        ('target_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_weights', numpy.empty(0, dtype=numpy.float32)),
    ])

    def __init__(self, onnx_node, desc=None, runtime_version=1, **options):
        TreeEnsembleRegressorCommon.__init__(
            self, numpy.float32, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleRegressor.atts,
            runtime_version=runtime_version, **options)


class TreeEnsembleRegressorDouble(TreeEnsembleRegressorCommon):

    atts = OrderedDict([
        ('aggregate_function', b'SUM'),
        ('base_values', numpy.empty(0, dtype=numpy.float64)),
        ('n_targets', 1),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float64)),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float64)),
        ('post_transform', b'NONE'),
        ('target_ids', numpy.empty(0, dtype=numpy.int64)),
        ('target_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_weights', numpy.empty(0, dtype=numpy.float64)),
    ])

    def __init__(self, onnx_node, desc=None, runtime_version=1, **options):
        TreeEnsembleRegressorCommon.__init__(
            self, numpy.float64, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleRegressorDouble.atts,
            runtime_version=runtime_version, **options)


class TreeEnsembleRegressorDoubleSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleRegressorDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'TreeEnsembleRegressorDouble')
        self.attributes = TreeEnsembleRegressorDouble.atts



"""
@file
@brief Command line about validation of prediction runtime.
"""
import json
from logging import getLogger
from ..asv_benchmark import create_asv_benchmark


def asv_bench(location='asvsklonnx', opset_min=-1, opset_max=None,
              runtime='scikit-learn,python_compiled', models=None,
              skip_models=None, extended_list=True,
              dims='1,10,100,1000,10000,100000',
              n_features='4,20', dtype=None,
              verbose=1, fLOG=print, clean=True, flat=False,
              conf_params=None, build=None, add_pyspy=False,
              env=None, matrix=None):
    """
    Creates an :epkg:`asv` benchmark in a folder
    but does not run it.

    :param location: location of the benchmark
    :param n_features: number of features to try
    :param dims: number of observations to try
    :param verbose: integer from 0 (None) to 2 (full verbose)
    :param opset_min: tries every conversion from this minimum opset,
        `-1` to get the current opset defined by module onnx
    :param opset_max: tries every conversion up to maximum opset,
        `-1` to get the current opset defined by module onnx
    :param runtime: runtime to check, *scikit-learn*, *python*,
        *python_compiled* compiles the graph structure
        and is more efficient when the number of observations is
        small, *onnxruntime1* to check :epkg:`onnxruntime`,
        *onnxruntime2* to check every ONNX node independently
        with onnxruntime, many runtime can be checked at the same time
        if the value is a comma separated list
    :param models: list of models to test or empty
        string to test them all
    :param skip_models: models to skip
    :param extended_list: extends the list of :epkg:`scikit-learn` converters
        with converters implemented in this module
    :param dtype: '32' or '64' or None for both,
        limits the test to one specific number types
    :param fLOG: logging function
    :param clean: clean the folder first, otherwise overwrites the content
    :param conf_params: to overwrite some of the configuration parameters,
        format ``name,value;name2,value2``
    :param flat: one folder for all files or subfolders
    :param build: location of the outputs (env, html, results)
    :param add_pyspy: add an extra folder with code to profile
        each configuration
    :param env: default environment or ``same`` to use the current one
    :param matrix: specifies versions for a module as a json string,
        example: ``{'onnxruntime': ['1.1.1', '1.1.2']}``,
        if a package name starts with `'~'`, the package is removed
    :return: created files

    .. cmdref::
        :title: Automatically creates an asv benchmark
        :cmd: -m mlprodict asv_bench --help
        :lid: l-cmd-asv-bench

        The command creates a benchmark based on asv module.
        It does not run it.

        Example::

            python -m mlprodict asv_bench --models LogisticRegression,LinearRegression
    """
    if not isinstance(models, list):
        models = (None if models in (None, "")
                  else models.strip().split(','))
    if not isinstance(skip_models, list):
        skip_models = ({} if skip_models in (None, "")
                       else skip_models.strip().split(','))
    if opset_max == "":
        opset_max = None  # pragma: no cover
    if isinstance(opset_min, str):
        opset_min = int(opset_min)  # pragma: no cover
    if isinstance(opset_max, str):
        opset_max = int(opset_max)  # pragma: no cover
    if isinstance(verbose, str):
        verbose = int(verbose)  # pragma: no cover
    if isinstance(extended_list, str):
        extended_list = extended_list in (
            '1', 'True', 'true')  # pragma: no cover
    if isinstance(add_pyspy, str):
        add_pyspy = add_pyspy in ('1', 'True', 'true')  # pragma: no cover
    if not isinstance(runtime, list):
        runtime = runtime.split(',')
    if not isinstance(dims, list):
        dims = [int(_) for _ in dims.split(',')]
    if matrix is not None:
        if matrix in ('None', ''):
            matrix = None  # pragma: no cover
        elif not isinstance(matrix, dict):
            matrix = json.loads(matrix)
    if not isinstance(n_features, list):
        if n_features in (None, ""):
            n_features = None  # pragma: no cover
        else:
            n_features = list(map(int, n_features.split(',')))
    flat = flat in (True, 'True', 1, '1')

    def fct_filter_exp(m, s):
        return str(m) not in skip_models

    if dtype in ('', None):
        fct_filter = fct_filter_exp
    elif dtype == '32':  # pragma: no cover
        def fct_filter_exp2(m, p):
            return fct_filter_exp(m, p) and '64' not in p
        fct_filter = fct_filter_exp2
    elif dtype == '64':
        def fct_filter_exp3(m, p):
            return fct_filter_exp(m, p) and '64' in p
        fct_filter = fct_filter_exp3
    else:
        raise ValueError(  # pragma: no cover
            "dtype must be empty, 32, 64 not '{}'.".format(dtype))

    if conf_params is not None:
        res = {}
        kvs = conf_params.split(';')
        for kv in kvs:
            spl = kv.split(',')
            if len(spl) != 2:
                raise ValueError(  # pragma: no cover
                    "Unable to interpret '{}'.".format(kv))
            k, v = spl
            res[k] = v
        conf_params = res

    if verbose <= 1:
        logger = getLogger('skl2onnx')
        logger.disabled = True

    return create_asv_benchmark(
        location=location, opset_min=opset_min, opset_max=opset_max,
        runtime=runtime, models=models, skip_models=skip_models,
        extended_list=extended_list, dims=dims,
        n_features=n_features, dtype=dtype, verbose=verbose,
        fLOG=fLOG, clean=clean, conf_params=conf_params,
        filter_exp=fct_filter, filter_scenario=None,
        flat=flat, build=build, add_pyspy=add_pyspy,
        env=env, matrix=matrix)



"""
@file
@brief Functions to help visualizing performances.
"""
import numpy
import pandas


def _model_name(name):
    """
    Extracts the main component of a model, removes
    suffixes such ``Classifier``, ``Regressor``, ``CV``.

    @param      name        string
    @return                 shorter string
    """
    if name.startswith("Select"):
        return "Select"
    if name.startswith("Nu"):
        return "Nu"
    modif = 1
    while modif > 0:
        modif = 0
        for suf in ['Classifier', 'Regressor', 'CV', 'IC',
                    'Transformer']:
            if name.endswith(suf):
                name = name[:-len(suf)]
                modif += 1
    return name


def plot_validate_benchmark(df):
    """
    Plots a graph which summarizes the performances of a benchmark
    validating a runtime for :epkg:`ONNX`.

    @param      df      output of function @see fn summary_report
    @return             fig, ax

    .. plot::

        from logging import getLogger
        from pandas import DataFrame
        import matplotlib.pyplot as plt
        from mlprodict.onnxrt.validate import enumerate_validated_operator_opsets, summary_report
        from mlprodict.tools.plotting import plot_validate_benchmark

        logger = getLogger('skl2onnx')
        logger.disabled = True

        rows = list(enumerate_validated_operator_opsets(
            verbose=0, models={"LinearRegression"}, opset_min=11,
            runtime=['python', 'onnxruntime1'], debug=False,
            benchmark=True, n_features=[None, 10]))

        df = DataFrame(rows)
        piv = summary_report(df)
        fig, ax = plot_validate_benchmark(piv)
        plt.show()
    """
    import matplotlib.pyplot as plt

    if 'n_features' not in df.columns:
        df["n_features"] = numpy.nan  # pragma: no cover
    if 'runtime' not in df.columns:
        df['runtime'] = '?'  # pragma: no cover

    fmt = "{} [{}-{}|{}] D{}"
    df["label"] = df.apply(
        lambda row: fmt.format(
            row["name"], row["problem"], row["scenario"],
            row['optim'], row["n_features"]).replace("-default|", "-**]"), axis=1)
    df = df.sort_values(["name", "problem", "scenario", "optim",
                         "n_features", "runtime"],
                        ascending=False).reset_index(drop=True).copy()
    indices = ['label', 'runtime']
    values = [c for c in df.columns
              if 'N=' in c and '-min' not in c and '-max' not in c]
    try:
        df = df[indices + values]
    except KeyError as e:  # pragma: no cover
        raise RuntimeError(
            "Unable to find the following columns {}\nin {}".format(
                indices + values, df.columns)) from e

    if 'RT/SKL-N=1' not in df.columns:
        raise RuntimeError(  # pragma: no cover
            "Column 'RT/SKL-N=1' is missing, benchmark was probably not run.")
    na = df["RT/SKL-N=1"].isnull()
    dfp = df[~na]
    runtimes = list(sorted(set(dfp['runtime'])))
    final = None
    for rt in runtimes:
        sub = dfp[dfp.runtime == rt].drop('runtime', axis=1).copy()
        col = list(sub.columns)
        for i in range(1, len(col)):
            col[i] += "__" + rt
        sub.columns = col

        if final is None:
            final = sub
        else:
            final = final.merge(sub, on='label', how='outer')

    # let's add average and median
    ncol = (final.shape[1] - 1) // len(runtimes)
    if len(runtimes) + 1 > final.shape[0]:
        dfp_legend = final.iloc[:len(runtimes) + 1, :].copy()
        while dfp_legend.shape[0] < len(runtimes) + 1:
            dfp_legend = pandas.concat([dfp_legend, dfp_legend[:1]])
    else:
        dfp_legend = final.iloc[:len(runtimes) + 1, :].copy()
    rleg = dfp_legend.copy()
    dfp_legend.iloc[:, 1:] = numpy.nan
    rleg.iloc[:, 1:] = numpy.nan

    for r, runt in enumerate(runtimes):
        sli = slice(1 + ncol * r, 1 + ncol * r + ncol)
        cm = final.iloc[:, sli].mean().values
        dfp_legend.iloc[r + 1, sli] = cm
        rleg.iloc[r, sli] = final.iloc[:, sli].median()
        dfp_legend.iloc[r + 1, 0] = "avg_" + runt
        rleg.iloc[r, 0] = "med_" + runt
    dfp_legend.iloc[0, 0] = "------"
    rleg.iloc[-1, 0] = "------"

    # sort
    final = final.sort_values('label', ascending=False).copy()

    # add global statistics
    final = pandas.concat([rleg, final, dfp_legend]).reset_index(drop=True)

    # graph beginning
    total = final.shape[0] * 0.45
    fig, ax = plt.subplots(1, len(values), figsize=(14, total),
                           sharex=False, sharey=True)
    x = numpy.arange(final.shape[0])
    subh = 1.0 / len(runtimes)
    height = total / final.shape[0] * (subh + 0.1)
    decrt = {rt: height * i for i, rt in enumerate(runtimes)}
    colors = {rt: c for rt, c in zip(
        runtimes, ['blue', 'orange', 'cyan', 'yellow'])}

    # draw lines between models
    vals = final.iloc[:, 1:].values.ravel()
    xlim = [min(0.5, min(vals)), max(2, max(vals))]
    while i < final.shape[0] - 1:
        i += 1
        label = final.iloc[i, 0]
        if '[' not in label:
            continue
        prev = final.iloc[i - 1, 0]
        if '[' not in label:
            continue  # pragma: no cover
        label = label.split()[0]
        prev = prev.split()[0]
        if _model_name(label) == _model_name(prev):
            continue

        blank = final.iloc[:1, :].copy()
        blank.iloc[0, 0] = '------'
        blank.iloc[0, 1:] = xlim[0]
        final = pandas.concat([final[:i], blank, final[i:]])
        i += 1

    final = final.reset_index(drop=True).copy()
    x = numpy.arange(final.shape[0])

    done = set()
    for c in final.columns[1:]:
        place, runtime = c.split('__')
        if hasattr(ax, 'shape'):
            index = values.index(place)
            if (index, runtime) in done:
                raise RuntimeError(  # pragma: no cover
                    "Issue with column '{}'\nlabels={}\nruntimes={}\ncolumns="
                    "{}\nvalues={}\n{}".format(
                        c, list(final.label), runtimes, final.columns, values, final))
            axi = ax[index]
            done.add((index, runtime))
        else:
            if (0, runtime) in done:  # pragma: no cover
                raise RuntimeError(
                    "Issue with column '{}'\nlabels={}\nruntimes={}\ncolumns="
                    "{}\nvalues={}\n{}".format(
                        c, final.label, runtimes, final.columns, values, final))
            done.add((0, runtime))  # pragma: no cover
            axi = ax  # pragma: no cover
        if c in final.columns:
            yl = final.loc[:, c]
            xl = x + decrt[runtime] / 2
            axi.barh(xl, yl, label=runtime, height=height,
                     color=colors[runtime])
            axi.set_title(place)

    def _plot_axis(axi, x, xlim):
        axi.plot([1, 1], [0, max(x)], 'g-')
        axi.plot([2, 2], [0, max(x)], 'r--')
        axi.set_xlim(xlim)
        axi.set_xscale('log')
        axi.set_ylim([min(x) - 2, max(x) + 1])

    def _plot_final(axi, x, final):
        axi.set_yticks(x)
        axi.set_yticklabels(final['label'])

    if hasattr(ax, 'shape'):
        for i in range(len(ax)):  # pylint: disable=C0200
            _plot_axis(ax[i], x, xlim)

        ax[min(ax.shape[0] - 1, 2)].legend()
        _plot_final(ax[0], x, final)
    else:  # pragma: no cover
        _plot_axis(ax, x, xlim)
        _plot_final(ax, x, final)
        ax.legend()

    fig.subplots_adjust(left=0.25)
    return fig, ax



"""
@file
@brief Action definition.
"""
import numpy
from .gactions import MLActionFunctionCall


class MLActionTensorDot(MLActionFunctionCall):
    """
    Scalar product.
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first tensor
        @param  act2    second tensor
        """
        MLActionFunctionCall.__init__(self, "adot", act1.output, act1, act2)
        # dot product takes two vectors and returns a float
        self.output = act1.output.element_type

    def _optional_parameters(self):
        return str(self.inputs[0].dim[0])

    def execute(self, **kwargs):
        """
        Addition
        """
        MLActionFunctionCall.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(self.output.softcast(numpy.dot(res[0], res[1])))


class MLActionTensorTake(MLActionFunctionCall):
    """
    Extracts an element of the tensor.
    """

    def __init__(self, tens, ind):
        """
        @param  tens    tensor
        @param  ind     index
        """
        MLActionFunctionCall.__init__(self, "atake", tens.output, tens, ind)
        self.output = tens.output.element_type

    def _optional_parameters(self):
        return str(self.inputs[0].dim[0])

    def execute(self, **kwargs):
        """
        Addition
        """
        MLActionFunctionCall.execute(self, **kwargs)
        res = self.ChildrenResults
        if res[1] < 0:
            raise ValueError(  # pragma: no cover
                "Cannot take element {0}".format(res[1]))
        if res[1] >= len(res[0]):
            raise ValueError(  # pragma: no cover
                "Cannot take element {0} >= size={1}".format(res[1], len(res[0])))
        return self.output.validate(self.output.softcast(res[0][res[1]]))


class MLActionTensorVector(MLActionFunctionCall):
    """
    Tensor operation.
    """

    def __init__(self, act1, act2, name, fct):
        """
        @param  act1    first tensor
        @param  act2    second tensor
        @param  name    operator name
        @param  fct     function
        """
        MLActionFunctionCall.__init__(self, name, act1.output, act1, act2)
        self.output = act1.output
        self.fct = fct

    def _optional_parameters(self):
        return str(self.inputs[0].dim[0])

    def execute(self, **kwargs):
        """
        Addition
        """
        MLActionFunctionCall.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(self.fct(res[0], res[1]))


class MLActionTensorSub(MLActionTensorVector):
    """
    Tensor soustraction.
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first tensor
        @param  act2    second tensor
        """
        MLActionTensorVector.__init__(
            self, act1, act2, "asub", lambda v1, v2: v1 - v2)


class MLActionTensorMul(MLActionTensorVector):
    """
    Tensor multiplication.
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first tensor
        @param  act2    second tensor
        """
        MLActionTensorVector.__init__(  # pragma: no cover
            self, act1, act2, "amul", lambda v1, v2: numpy.multiply(v1, v2))


class MLActionTensorDiv(MLActionTensorVector):
    """
    Tensor division.
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first tensor
        @param  act2    second tensor
        """
        MLActionTensorVector.__init__(
            self, act1, act2, "adiv", lambda v1, v2: numpy.divide(v1, v2))


class MLActionTensorAdd(MLActionTensorVector):
    """
    Tensor addition.
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first tensor
        @param  act2    second tensor
        """
        MLActionTensorVector.__init__(  # pragma: no cover
            self, act1, act2, "aadd", lambda v1, v2: v1 + v2)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def gather_numpy_2(self, dim, index):
    res = []
    for a, b in zip(self, index):
        res.append(a[b[0]])
    res = numpy.array(
        res, dtype=self.dtype).reshape(
            index.shape)
    return res


def gather_numpy(self, dim, index):
    """
    Gathers values along an axis specified by dim.
    For a 3-D tensor the output is specified by:

    ::

        out[i][j][k] = input[index[i][j][k]][j][k]  # if dim == 0
        out[i][j][k] = input[i][index[i][j][k]][k]  # if dim == 1
        out[i][j][k] = input[i][j][index[i][j][k]]  # if dim == 2

    :param dim: The axis along which to index
    :param index: A tensor of indices of elements to gather
    :return: tensor of gathered values

    See `How to do scatter and gather operations in numpy?
    <https://stackoverflow.com/questions/46065873/
    how-to-do-scatter-and-gather-operations-in-numpy/46204790#46204790>`_
    """
    idx_xsection_shape = index.shape[:dim] + index.shape[dim + 1:]
    self_xsection_shape = self.shape[:dim] + self.shape[dim + 1:]
    if idx_xsection_shape != self_xsection_shape:
        raise ValueError(  # pragma: no cover
            "Except for dimension {}, all dimensions of "
            "index and self should be the same size".format(dim))
    data_swaped = numpy.swapaxes(self, 0, dim)
    index_swaped = numpy.swapaxes(index, 0, dim)
    try:
        gathered = numpy.choose(index_swaped, data_swaped)
    except ValueError as e:
        if len(index_swaped.shape) == 2 and len(data_swaped.shape) == 2:
            return gather_numpy_2(self, dim, index)
        raise e  # pragma: no cover

    return numpy.swapaxes(gathered, 0, dim)


class GatherElements(OpRun):

    atts = {'axis': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=GatherElements.atts,
                       **options)

    def _run(self, data, indices):  # pylint: disable=W0221
        y = gather_numpy(data, self.axis, indices)
        return (y, )

    def _infer_shapes(self, data, indices):  # pylint: disable=W0221
        return (indices, )

    def to_python(self, inputs):
        lines = ['data_swaped = numpy.swapaxes(%s, 0, axis)' % inputs[0],
                 'index_swaped = numpy.swapaxes(%s, 0, axis)' % inputs[1],
                 "gathered = numpy.choose(index_swaped, data_swaped, mode='wrap')",
                 'return numpy.swapaxes(gathered, 0, axis)']
        return "import numpy", "\n".join(lines)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class CumSum(OpRun):

    atts = {'exclusive': 0, 'reverse': 0}
    python_inputs = ['x', 'axis=None']

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=CumSum.atts,
                       **options)

    def _run(self, x, *axis):  # pylint: disable=W0221
        axis = None if len(axis) == 0 else axis[0]
        if axis is None:
            if self.reverse or self.exclusive:
                raise NotImplementedError(  # pragma no cover
                    'reverse=1 or exclusive=1 not implemented')
            if self.inplaces.get(0, False):
                return (numpy.cumsum(x, out=x), )
            return (numpy.cumsum(x), )
        if not isinstance(axis, (numpy.int32, numpy.int64)):
            if (len(axis.shape) > 1 or
                    (len(axis.shape) > 0 and axis.shape[0] != 1)):
                raise RuntimeError(  # pragma no cover
                    "axis must be an array of one number not {} "
                    "(shape {})".format(axis, axis.shape))
            if len(axis.shape) > 0:
                axis = axis[0]
        if self.reverse or self.exclusive:
            raise NotImplementedError(
                'reverse=1 or exclusive=1 not implemented')
        if self.inplaces.get(0, False):
            return (numpy.cumsum(x, axis=axis, out=x), )
        return (numpy.cumsum(x, axis=axis), )

    def _infer_shapes(self, x, *axis):  # pylint: disable=W0221
        return (x, )

    def to_python(self, inputs):
        lines = ['if exclusive or reverse:',
                 '    raise NotImplementedError("reverse=1 or exclusive=1 not implemente")',
                 'if axis is None:',
                 '    return numpy.cumsum(x)',
                 'return numpy.cumsum(x, axis=axis[0])']
        return 'import numpy', "\n".join(lines)



# -*- encoding: utf-8 -*-
# pylint: disable=W0611
"""
@file
@brief Imports runtime operators.
"""

from ._op import OpRun
from .op_abs import Abs
from .op_add import Add
from .op_and import And
from .op_argmax import ArgMax
from .op_argmin import ArgMin
from .op_array_feature_extractor import ArrayFeatureExtractor
from .op_atan import Atan
from .op_batch_normalization import BatchNormalization
from .op_binarizer import Binarizer
from .op_cast import Cast
from .op_cdist import CDist
from .op_ceil import Ceil
from .op_celu import Celu
from .op_clip import Clip_6, Clip_11, Clip
from .op_concat import Concat
from .op_conv import Conv
from .op_conv_transpose import ConvTranspose
from .op_constant import Constant
from .op_constant_of_shape import ConstantOfShape
from .op_cum_sum import CumSum
from .op_dequantize_linear import DequantizeLinear
from .op_dict_vectorizer import DictVectorizer
from .op_div import Div
from .op_dropout import Dropout, Dropout_7, Dropout_12
from .op_einsum import Einsum
from .op_equal import Equal
from .op_erf import Erf
from .op_exp import Exp
from .op_eyelike import EyeLike
from .op_feature_vectorizer import FeatureVectorizer
from .op_flatten import Flatten
from .op_gather import Gather
from .op_gather_elements import GatherElements
from .op_gemm import Gemm
from .op_global_average_pool import GlobalAveragePool
from .op_greater import Greater, GreaterOrEqual
from .op_floor import Floor
from .op_identity import Identity
from .op_if import If
from .op_imputer import Imputer
from .op_isnan import IsNaN
from .op_label_encoder import LabelEncoder
from .op_less import Less, LessOrEqual
from .op_linear_classifier import LinearClassifier
from .op_linear_regressor import LinearRegressor
from .op_log import Log
from .op_loop import Loop
from .op_lp_normalization import LpNormalization
from .op_matmul import MatMul
from .op_max import Max
from .op_max_pool import MaxPool
from .op_mean import Mean
from .op_min import Min
from .op_mul import Mul
from .op_neg import Neg
from .op_normalizer import Normalizer
from .op_not import Not
from .op_one_hot_encoder import OneHotEncoder
from .op_or import Or
from .op_pad import Pad
from .op_pow import Pow
from .op_quantize_linear import QuantizeLinear
from .op_reciprocal import Reciprocal
from .op_reduce_log_sum_exp import ReduceLogSumExp
from .op_reduce_min import ReduceMin
from .op_reduce_max import ReduceMax
from .op_reduce_mean import ReduceMean
from .op_reduce_prod import ReduceProd
from .op_reduce_sum import (
    ReduceSum_1, ReduceSum_11, ReduceSum_13, ReduceSum)
from .op_reduce_sum_square import ReduceSumSquare
from .op_relu import Relu
from .op_reshape import Reshape
from .op_rnn import RNN
from .op_scaler import Scaler
from .op_scan import Scan
from .op_shape import Shape
from .op_sigmoid import Sigmoid
from .op_sign import Sign
from .op_sin import Sin
from .op_slice import Slice, Slice_1, Slice_10
from .op_split import Split, Split_2, Split_11, Split_13
from .op_softmax import Softmax
from .op_solve import Solve
from .op_sqrt import Sqrt
from .op_squeeze import Squeeze, Squeeze_1, Squeeze_11, Squeeze_13
from .op_string_normalizer import StringNormalizer
from .op_sub import Sub
from .op_sum import Sum
from .op_svm_classifier import SVMClassifier, SVMClassifierDouble
from .op_svm_regressor import SVMRegressor, SVMRegressorDouble
from .op_tfidfvectorizer import TfIdfVectorizer
from .op_tokenizer import Tokenizer
from .op_topk import TopK_10, TopK_11, TopK_1, TopK
from .op_transpose import Transpose
from .op_tree_ensemble_classifier import TreeEnsembleClassifier, TreeEnsembleClassifierDouble
from .op_tree_ensemble_regressor import TreeEnsembleRegressor, TreeEnsembleRegressorDouble
from .op_unsqueeze import Unsqueeze, Unsqueeze_1, Unsqueeze_11, Unsqueeze_13
from .op_where import Where
from .op_zipmap import ZipMap


from ..doc.doc_helper import get_rst_doc
_op_list = []
clo = locals().copy()
for name, cl in clo.items():
    if not cl.__doc__ and issubclass(cl, OpRun):
        cl.__doc__ = get_rst_doc(cl.__name__)
        _op_list.append(cl)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from onnx.defs import onnx_opset_version
from ..shape_object import ShapeObject
from ._op import OpRun


class SliceCommon(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, data, starts, ends, axes=None, steps=None):  # pylint: disable=W0221
        if axes is None:
            if steps is None:
                slices = [slice(s, e) for s, e in zip(starts, ends)]
            else:
                slices = [slice(s, e, d)
                          for s, e, d in zip(starts, ends, steps)]
        else:
            if steps is None:
                slices = [slice(0, a) for a in data.shape]
                for s, e, a in zip(starts, ends, axes):
                    slices[a] = slice(s, e)
            else:
                slices = [slice(0, a) for a in data.shape]
                for s, e, a, d in zip(starts, ends, axes, steps):
                    slices[a] = slice(s, e, d)
        return (data[tuple(slices)], )

    def _infer_shapes(self, data, starts, ends, axes=None, steps=None):  # pylint: disable=W0221
        pref = str(hex(id(self))[2:])
        shape = ["nslice%s_%d" % (pref, i) for i in range(len(data.shape))]
        return (ShapeObject(shape, data.dtype), )


class Slice_10(SliceCommon):
    def __init__(self, onnx_node, desc=None, **options):
        SliceCommon.__init__(self, onnx_node, desc=desc,
                             **options)


class Slice_1(SliceCommon):

    atts = {'starts': [], 'ends': [], 'axes': []}

    def __init__(self, onnx_node, desc=None, **options):
        SliceCommon.__init__(self, onnx_node, desc=desc,
                             expected_attributes=Slice_1.atts,
                             **options)
        for f in ['starts', 'ends', 'steps', 'axes']:
            if not hasattr(self, f):
                continue
            if getattr(self, f) is not None and len(getattr(self, f)) == 0:
                setattr(self, f, None)

    def _run(self, data):  # pylint: disable=W0221
        return SliceCommon._run(
            self, data, self.starts, self.ends, self.axes)

    def _infer_shapes(self, data):  # pylint: disable=W0221
        return SliceCommon._infer_shapes(
            self, data, self.starts, self.ends, self.axes)


if onnx_opset_version() >= 10:
    Slice = Slice_10
else:
    Slice = Slice_1  # pragma: no cover



"""
@file
@brief Custom operator Tokenizer.
"""
from skl2onnx.algebra.onnx_operator import OnnxOperator


class OnnxTokenizer_1(OnnxOperator):
    """
    Defines a custom operator not defined by ONNX
    specifications but in onnxruntime.
    """

    since_version = 1
    expected_inputs = ['text']
    expected_outputs = ['tokens']
    input_range = [1, 1]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'mlprodict'
    operator_name = 'Tokenizer'
    past_version = {}

    def __init__(self, text, mark=0, mincharnum=1,
                 pad_value='#', separators=None,
                 tokenexp='[a-zA-Z0-9_]+', stopwords=None,
                 op_version=None, **kwargs):
        """
        :param text: array or OnnxOperatorMixin
        :param mark: see :epkg:`Tokenizer`
        :param pad_value: see :epkg:`Tokenizer`
        :param separators: see :epkg:`Tokenizer`
        :param tokenexp: see :epkg:`Tokenizer`
        :param stopwords: list of stopwords, addition to :epkg:`Tokenizer`
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        if separators is None:
            separators = []
        if stopwords is None:
            stopwords = []
        OnnxOperator.__init__(
            self, text, mark=mark, mincharnum=mincharnum,
            pad_value=pad_value, separators=separators, tokenexp=tokenexp,
            stopwords=stopwords, op_version=op_version, **kwargs)


OnnxTokenizer = OnnxTokenizer_1



# -*- coding: utf-8 -*-
"""
@file
@brief Converters from scikit-learn model.
"""
import numpy
from .g_sklearn_type_helpers import check_type
from .grammar.gactions import MLActionVar, MLActionCst, MLActionReturn
from .grammar.gactions_tensor import MLActionTensorDiv, MLActionTensorSub
from .grammar.gmlactions import MLModel


def sklearn_standard_scaler(model, input_names=None, output_names=None, **kwargs):
    """
    Converts a `standard scaler <http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html>`_
    model into a *grammar* model (semantic graph representation).

    @param      model           scikit-learn model
    @param      input_names     name of the input features
    @param      output_names    name of the output predictions
    @param      kwargs          additional parameters (none)
    @return                     graph model

    If *input* is None or *output* is None, default values
    will be given to the outputs
    ``['Prediction', 'Score']`` for the outputs.
    If *input_names* is None, it wil be ``'Features'``.

    No additional parameters is considered.
    """
    if output_names is None:
        output_names = ['Prediction', 'Score']
    if input_names is None:
        input_names = 'Features'

    from sklearn.preprocessing import StandardScaler
    check_type(model, StandardScaler)

    lmean = MLActionCst(model.mean_.ravel().astype(numpy.float32))
    lscale = MLActionCst(model.scale_.ravel().astype(numpy.float32))

    lvar = MLActionVar(model.var_.astype(numpy.float32), input_names)
    lno = MLActionTensorSub(lvar, lmean)
    lno = MLActionTensorDiv(lno, lscale)
    ret = MLActionReturn(lno)
    return MLModel(ret, output_names, name=StandardScaler.__name__)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject


def _pad_impl(data, raw_pads, mode, constant_values=0.0):
    input_rank = data.ndim
    if input_rank * 2 != raw_pads.size:
        raise RuntimeError(  # pragma: no cover
            'The number of elements in raw_pads should be 2 * data_rank')

    half = raw_pads.shape[0] // 2
    pad_width = tuple((raw_pads[i], raw_pads[i + half])
                      for i in range(0, half))

    if mode == 'constant':
        return numpy.pad(data, pad_width=pad_width, mode=mode,
                         constant_values=constant_values)
    return numpy.pad(data, pad_width=pad_width, mode=mode)


class Pad(OpRun):

    atts = {'mode': b'constant'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Pad.atts,
                       **options)
        self.mode_ = self.mode.decode('ascii')

    def _run(self, data, pads, constant_value=None):  # pylint: disable=W0221
        return (_pad_impl(data, pads, mode=self.mode_,
                          constant_values=constant_value), )

    def _infer_shapes(self, data, pads, constant_value=None):  # pylint: disable=E0202,W0221
        """
        Returns an empty shape by default.
        """
        return (ShapeObject(None, data.dtype), )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject


def reshape_reference_implementation(data, shape):
    new_shape = numpy.copy(shape)
    zeros_index = numpy.where(shape == 0)
    new_shape[zeros_index] = numpy.array(data.shape)[zeros_index]
    reshaped = numpy.reshape(data, new_shape)
    return reshaped


class Reshape(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, data, shape):  # pylint: disable=W0221
        return (reshape_reference_implementation(data, shape), )

    def _infer_shapes(self, data, shape):  # pylint: disable=W0221
        return (ShapeObject(None, dtype=data.dtype), )



"""
@file
@brief Inspired from :epkg:`sklearn-onnx`, handles two backends.
"""
import os
import pickle
import numpy
from numpy.testing import assert_array_almost_equal, assert_array_equal
from scipy.sparse.csr import csr_matrix
import pandas
from ...onnxrt.ops_cpu.op_zipmap import ArrayZipMapDictionary


class ExpectedAssertionError(Exception):
    """
    Expected failure.
    """
    pass


class OnnxBackendAssertionError(AssertionError):
    """
    Expected failure.
    """
    pass


class OnnxBackendMissingNewOnnxOperatorException(OnnxBackendAssertionError):
    """
    Raised when :epkg:`onnxruntime` or :epkg:`mlprodict`
    does not implement a new operator
    defined in the latest onnx.
    """
    pass


class OnnxRuntimeMissingNewOnnxOperatorException(OnnxBackendAssertionError):
    """
    Raised when a new operator was added but cannot be found.
    """
    pass


def evaluate_condition(backend, condition):
    """
    Evaluates a condition such as
    ``StrictVersion(onnxruntime.__version__) <= StrictVersion('0.1.3')``
    """
    if backend == "onnxruntime":  # pragma: no cover
        import onnxruntime  # pylint: disable=W0611
        return eval(condition)  # pylint: disable=W0123
    raise NotImplementedError(  # pragma no cover
        "Not implemented for backend '{0}' and "
        "condition '{1}'.".format(backend, condition))


def is_backend_enabled(backend):
    """
    Tells if a backend is enabled.
    Raises an exception if backend != 'onnxruntime'.
    Unit tests only test models against this backend.
    """
    if backend == "onnxruntime":
        try:
            import onnxruntime  # pylint: disable=W0611
            return True
        except ImportError:  # pragma no cover
            return False
    if backend == "python":
        return True
    raise NotImplementedError(  # pragma no cover
        "Not implemented for backend '{0}'".format(backend))


def load_data_and_model(items_as_dict, **context):
    """
    Loads every file in a dictionary {key: filename}.
    The extension is either *pkl* and *onnx* and determines
    how it it loaded. If the value is not a string,
    the function assumes it was already loaded.
    """
    res = {}
    for k, v in items_as_dict.items():
        if isinstance(v, str):
            if os.path.splitext(v)[-1] == ".pkl":
                with open(v, "rb") as f:  # pragma: no cover
                    try:
                        bin = pickle.load(f)
                    except ImportError as e:
                        if '.model.' in v:
                            continue
                        raise ImportError(  # pylint: disable=W0707
                            "Unable to load '{0}' due to {1}".format(v, e))
                    res[k] = bin
            else:
                res[k] = v
        else:
            res[k] = v
    return res


def extract_options(name):
    """
    Extracts comparison option from filename.
    As example, ``Binarizer-SkipDim1`` means
    options *SkipDim1* is enabled.
    ``(1, 2)`` and ``(2,)`` are considered equal.
    Available options: see :func:`dump_data_and_model`.
    """
    opts = name.replace("\\", "/").split("/")[-1].split('.')[0].split('-')
    if len(opts) == 1:
        return {}
    res = {}
    for opt in opts[1:]:
        if opt in ("SkipDim1", "OneOff", "NoProb", "NoProbOpp",
                   "Dec4", "Dec3", "Dec2", 'Svm',
                   'Out0', 'Reshape', 'SklCol', 'DF', 'OneOffArray'):
            res[opt] = True
        else:
            raise NameError("Unable to parse option '{}'".format(
                opts[1:]))  # pragma no cover
    return res


def compare_outputs(expected, output, verbose=False, **kwargs):
    """
    Compares expected values and output.
    Returns None if no error, an exception message otherwise.
    """
    SkipDim1 = kwargs.pop("SkipDim1", False)
    NoProb = kwargs.pop("NoProb", False)
    NoProbOpp = kwargs.pop("NoProbOpp", False)
    Dec4 = kwargs.pop("Dec4", False)
    Dec3 = kwargs.pop("Dec3", False)
    Dec2 = kwargs.pop("Dec2", False)
    Disc = kwargs.pop("Disc", False)
    Mism = kwargs.pop("Mism", False)

    if Dec4:
        kwargs["decimal"] = min(kwargs["decimal"], 4)
    if Dec3:
        kwargs["decimal"] = min(kwargs["decimal"], 3)
    if Dec2:
        kwargs["decimal"] = min(kwargs["decimal"], 2)  # pragma: no cover
    if isinstance(expected, numpy.ndarray) and isinstance(
            output, numpy.ndarray):
        if SkipDim1:
            # Arrays like (2, 1, 2, 3) becomes (2, 2, 3)
            # as one dimension is useless.
            expected = expected.reshape(
                tuple([d for d in expected.shape if d > 1]))
            output = output.reshape(tuple([d for d in expected.shape
                                           if d > 1]))
        if NoProb or NoProbOpp:
            # One vector is (N,) with scores, negative for class 0
            # positive for class 1
            # The other vector is (N, 2) score in two columns.
            if len(output.shape) == 2 and output.shape[1] == 2 and len(
                    expected.shape) == 1:
                output = output[:, 1]
                if NoProbOpp:
                    output = -output
            elif len(output.shape) == 1 and len(expected.shape) == 1:
                pass
            elif len(expected.shape) == 1 and len(output.shape) == 2 and \
                    expected.shape[0] == output.shape[0] and \
                    output.shape[1] == 1:
                output = output[:, 0]
                if NoProbOpp:
                    output = -output
            elif expected.shape != output.shape:
                raise NotImplementedError("Shape mismatch: {0} != {1}".format(  # pragma no cover
                    expected.shape, output.shape))
        if len(expected.shape) == 1 and len(
                output.shape) == 2 and output.shape[1] == 1:
            output = output.ravel()
        if len(output.shape) == 3 and output.shape[0] == 1 and len(
                expected.shape) == 2:
            output = output.reshape(output.shape[1:])
        if expected.dtype in (numpy.str, numpy.dtype("<U1"),
                              numpy.dtype("<U3")):
            try:
                assert_array_equal(expected, output, verbose=verbose)
            except Exception as e:  # pylint: disable=W0703
                if Disc:  # pragma no cover
                    # Bug to be fixed later.
                    return ExpectedAssertionError(str(e))
                else:  # pragma no cover
                    return OnnxBackendAssertionError(str(e))
        else:
            try:
                assert_array_almost_equal(expected,
                                          output,
                                          verbose=verbose,
                                          **kwargs)
            except (RuntimeError, AssertionError) as e:  # pragma no cover
                longer = "\n--EXPECTED--\n{0}\n--OUTPUT--\n{1}".format(
                    expected, output) if verbose else ""
                expected_ = numpy.asarray(expected).ravel()
                output_ = numpy.asarray(output).ravel()
                if len(expected_) == len(output_):
                    if numpy.issubdtype(expected_.dtype, numpy.floating):
                        diff = numpy.abs(expected_ - output_).max()
                    else:
                        diff = max((1 if ci != cj else 0)
                                   for ci, cj in zip(expected_, output_))
                    if diff == 0:
                        return None
                elif Mism:
                    return ExpectedAssertionError(
                        "dimension mismatch={0}, {1}\n{2}{3}".format(
                            expected.shape, output.shape, e, longer))
                else:
                    return OnnxBackendAssertionError(
                        "dimension mismatch={0}, {1}\n{2}{3}".format(
                            expected.shape, output.shape, e, longer))
                if Disc:
                    # Bug to be fixed later.
                    return ExpectedAssertionError(
                        "max-diff={0}\n--expected--output--\n{1}{2}".format(
                            diff, e, longer))
                return OnnxBackendAssertionError(
                    "max-diff={0}\n--expected--output--\n{1}{2}".format(
                        diff, e, longer))
    else:
        return OnnxBackendAssertionError(  # pragma: no cover
            "Unexpected types {0} != {1}".format(
                type(expected), type(output)))
    return None


def _post_process_output(res):
    """
    Applies post processings before running the comparison
    such as changing type from list to arrays.
    """
    if isinstance(res, list):
        if len(res) == 0:
            return res
        if len(res) == 1:
            return _post_process_output(res[0])
        if isinstance(res[0], numpy.ndarray):
            return numpy.array(res)
        if isinstance(res[0], dict):
            return pandas.DataFrame(res).values
        ls = [len(r) for r in res]
        mi = min(ls)
        if mi != max(ls):
            raise NotImplementedError(  # pragma no cover
                "Unable to postprocess various number of "
                "outputs in [{0}, {1}]"
                .format(min(ls), max(ls)))
        if mi > 1:
            output = []
            for i in range(mi):
                output.append(_post_process_output([r[i] for r in res]))
            return output
        if isinstance(res[0], list):
            # list of lists
            if isinstance(res[0][0], list):
                return numpy.array(res)
            if len(res[0]) == 1 and isinstance(res[0][0], dict):
                return _post_process_output([r[0] for r in res])
            if len(res) == 1:
                return res
            if len(res[0]) != 1:
                raise NotImplementedError(  # pragma no cover
                    "Not conversion implemented for {0}".format(res))
            st = [r[0] for r in res]
            return numpy.vstack(st)
        return res
    return res


def _create_column(values, dtype):
    "Creates a column from values with dtype"
    if str(dtype) == "tensor(int64)":
        return numpy.array(values, dtype=numpy.int64)
    if str(dtype) == "tensor(float)":
        return numpy.array(values, dtype=numpy.float32)
    if str(dtype) in ("tensor(double)", "tensor(float64)"):
        return numpy.array(values, dtype=numpy.float64)
    if str(dtype) in ("tensor(string)", "tensor(str)"):
        return numpy.array(values, dtype=numpy.str)
    raise OnnxBackendAssertionError(
        "Unable to create one column from dtype '{0}'".format(dtype))


def _compare_expected(expected, output, sess, onnx_model,
                      decimal=5, verbose=False, classes=None,
                      **kwargs):
    """
    Compares the expected output against the runtime outputs.
    This is specific to :epkg:`onnxruntime` or :epkg:`mlprodict`.
    """
    tested = 0
    if isinstance(expected, list):
        if isinstance(output, list):
            if 'Out0' in kwargs:
                expected = expected[:1]
                output = output[:1]
                del kwargs['Out0']
            if 'Reshape' in kwargs:
                del kwargs['Reshape']
                output = numpy.hstack(output).ravel()
                output = output.reshape(
                    (len(expected), len(output.ravel()) // len(expected)))
            if len(expected) != len(output):
                raise OnnxBackendAssertionError(  # pragma no cover
                    "Unexpected number of outputs '{0}', expected={1}, got={2}"
                    .format(onnx_model, len(expected), len(output)))
            for exp, out in zip(expected, output):
                _compare_expected(exp, out, sess, onnx_model, decimal=5, verbose=verbose,
                                  classes=classes, **kwargs)
                tested += 1
        else:
            raise OnnxBackendAssertionError(  # pragma no cover
                "Type mismatch for '{0}', output type is {1}".format(
                    onnx_model, type(output)))
    elif isinstance(expected, dict):
        if not isinstance(output, dict):
            raise OnnxBackendAssertionError(  # pragma no cover
                "Type mismatch for '{0}'".format(onnx_model))
        for k, v in output.items():
            if k not in expected:
                continue
            msg = compare_outputs(
                expected[k], v, decimal=decimal, verbose=verbose, **kwargs)
            if msg:
                raise OnnxBackendAssertionError(  # pragma no cover
                    "Unexpected output '{0}' in model '{1}'\n{2}".format(
                        k, onnx_model, msg))
            tested += 1
    elif isinstance(expected, numpy.ndarray):
        if isinstance(output, list):
            if expected.shape[0] == len(output) and isinstance(
                    output[0], dict):
                if isinstance(output, ArrayZipMapDictionary):
                    output = pandas.DataFrame(list(output))
                else:
                    output = pandas.DataFrame(output)
                output = output[list(sorted(output.columns))]
                output = output.values
        if isinstance(output, (dict, list)):
            if len(output) != 1:  # pragma: no cover
                ex = str(output)
                if len(ex) > 170:
                    ex = ex[:170] + "..."
                raise OnnxBackendAssertionError(
                    "More than one output when 1 is expected "
                    "for onnx '{0}'\n{1}"
                    .format(onnx_model, ex))
            output = output[-1]
        if not isinstance(output, numpy.ndarray):
            raise OnnxBackendAssertionError(  # pragma no cover
                "output must be an array for onnx '{0}' not {1}".format(
                    onnx_model, type(output)))
        if (classes is not None and (
                expected.dtype == numpy.str or expected.dtype.char == 'U')):
            try:
                output = numpy.array([classes[cl] for cl in output])
            except IndexError as e:  # pragma no cover
                raise RuntimeError('Unable to handle\n{}\n{}\n{}'.format(
                    expected, output, classes)) from e
        msg = compare_outputs(
            expected, output, decimal=decimal, verbose=verbose, **kwargs)
        if isinstance(msg, ExpectedAssertionError):
            raise msg  # pylint: disable=E0702
        if msg:
            raise OnnxBackendAssertionError(  # pragma no cover
                "Unexpected output in model '{0}'\n{1}".format(onnx_model, msg))
        tested += 1
    else:
        if isinstance(expected, csr_matrix):
            # DictVectorizer
            one_array = numpy.array(output)
            dense = numpy.asarray(expected.todense())
            msg = compare_outputs(dense, one_array, decimal=decimal,
                                  verbose=verbose, **kwargs)
            if msg:
                raise OnnxBackendAssertionError(  # pragma no cover
                    "Unexpected output in model '{0}'\n{1}".format(onnx_model, msg))
            tested += 1
        else:
            raise OnnxBackendAssertionError(  # pragma no cover
                "Unexpected type for expected output ({1}) and onnx '{0}'".
                format(onnx_model, type(expected)))
    if tested == 0:
        raise OnnxBackendAssertionError(  # pragma no cover
            "No test for onnx '{0}'".format(onnx_model))



"""
@file
@brief Validates runtime for many :scikit-learn: operators.
The submodule relies on :epkg:`onnxconverter_common`,
:epkg:`sklearn-onnx`.
"""
import numpy
import pandas


def measure_relative_difference(skl_pred, ort_pred, batch=True):
    """
    Measures the relative difference between predictions
    between two ways of computing them.
    The functions returns nan if shapes are different.

    @param      skl_pred        prediction from :epkg:`scikit-learn`
                                or any other way
    @param      ort_pred        prediction from an :epkg:`ONNX` runtime
                                or any other way
    @param      batch           predictions are processed in a batch,
                                *skl_pred* and *ort_pred* should be arrays
                                or tuple or list of arrays
    @return                     relative max difference
                                or nan if it does not make any sense

    Because approximations get bigger when the vector is high,
    the function computes an adjusted relative differences.
    Let's assume *X* and *Y* are two vectors, let's denote
    :math:`med(X)` the median of *X*. The function returns the
    following metric: :math:`\\max_i(|X_i - Y_i| / \\max(X_i, med(|X|))`.

    The function takes the fourth highest difference, not the three first
    which may happen after a conversion into float32.
    """
    if hasattr(ort_pred, "is_zip_map") and ort_pred.is_zip_map:
        ort_pred = ort_pred.values
    if (isinstance(skl_pred, list) and
            all(map(lambda t: isinstance(t, numpy.ndarray), skl_pred))):
        # multi label classification
        skl_pred = numpy.array(skl_pred)
        skl_pred = skl_pred.reshape((skl_pred.shape[1], -1))

    if isinstance(skl_pred, tuple) or (batch and isinstance(skl_pred, list)):
        diffs = []
        if batch:
            if len(skl_pred) != len(ort_pred):
                return 1e10  # pragma: no cover
            for i in range(len(skl_pred)):  # pylint: disable=C0200
                diff = measure_relative_difference(skl_pred[i], ort_pred[i])
                diffs.append(diff)
        else:  # pragma: no cover
            for i in range(len(skl_pred)):  # pylint: disable=C0200
                try:
                    diff = measure_relative_difference(
                        skl_pred[i], [_[i] for _ in ort_pred])
                except IndexError:  # pragma: no cover
                    return 1e9
                except RuntimeError as e:  # pragma: no cover
                    raise RuntimeError("Unable to compute differences between"
                                       "\n{}--------\n{}".format(
                                           skl_pred, ort_pred)) from e
                diffs.append(diff)
        return max(diffs)
    else:
        ort_pred_ = ort_pred
        if isinstance(ort_pred, list):
            if isinstance(ort_pred[0], dict):
                ort_pred = pandas.DataFrame(list(ort_pred)).values
            elif (isinstance(ort_pred[0], list) and
                    isinstance(ort_pred[0][0], dict)):
                if len(ort_pred) == 1:  # pragma: no cover
                    ort_pred = pandas.DataFrame(list(ort_pred[0])).values
                elif len(ort_pred[0]) == 1:  # pragma: no cover
                    ort_pred = pandas.DataFrame(
                        [o[0] for o in ort_pred]).values
                else:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to compute differences between"
                        "\n{}--------\n{}".format(skl_pred, ort_pred))
            else:
                try:
                    ort_pred = numpy.array(ort_pred)
                except ValueError as e:  # pragma: no cover
                    raise ValueError(
                        "Unable to interpret (batch={}, type(skl_pred): {})\n{}\n-----\n{}".format(
                            batch, type(skl_pred), skl_pred, ort_pred)) from e

        if hasattr(skl_pred, 'todense'):
            skl_pred = skl_pred.todense().getA()
            skl_sparse = True
        else:
            skl_sparse = False
        if hasattr(ort_pred, 'todense'):
            ort_pred = ort_pred.todense().getA()
            ort_sparse = True
        else:
            ort_sparse = False

        try:
            if (any(numpy.isnan(skl_pred.reshape((-1, )))) and
                    all(~numpy.isnan(ort_pred.reshape((-1, ))))):
                skl_pred = numpy.nan_to_num(skl_pred)
            if (any(numpy.isnan(ort_pred.reshape((-1, )))) and
                    all(~numpy.isnan(skl_pred.reshape((-1, ))))):
                ort_pred = numpy.nan_to_num(ort_pred)
        except ValueError as e:  # pragma: no cover
            raise RuntimeError(
                "Unable to compute differences between {}{} - {}{}\n{}\n{}\n"
                "--------\n{}".format(
                    skl_pred.shape, " (sparse)" if skl_sparse else "",
                    ort_pred.shape, " (sparse)" if ort_sparse else "",
                    e, skl_pred, ort_pred)) from e

        if isinstance(ort_pred, list):
            raise RuntimeError(  # pragma: no cover
                "Issue with {}\n{}".format(ort_pred, ort_pred_))

        if skl_pred.shape != ort_pred.shape and skl_pred.size == ort_pred.size:
            ort_pred = ort_pred.ravel()
            skl_pred = skl_pred.ravel()

        if skl_pred.shape != ort_pred.shape:
            return 1e11

        if hasattr(skl_pred, 'A'):
            # ravel() on matrix still returns a matrix
            skl_pred = skl_pred.A  # pragma: no cover
        if hasattr(ort_pred, 'A'):
            # ravel() on matrix still returns a matrix
            ort_pred = ort_pred.A  # pragma: no cover
        r_skl_pred = skl_pred.ravel()
        r_ort_pred = ort_pred.ravel()
        ab = numpy.abs(r_skl_pred)
        median = numpy.median(ab.ravel())
        mx = numpy.max(ab)
        if median == 0:
            median = mx
        if median == 0:
            median = 1
        mx = numpy.maximum(ab, median)
        d = (r_ort_pred - r_skl_pred) / mx
        rel_sort = numpy.sort(numpy.abs(d))
        rel_diff = rel_sort[-4] if len(rel_sort) > 5 else rel_sort[-1]

        if numpy.isnan(rel_diff) and not all(numpy.isnan(r_ort_pred)):
            raise RuntimeError(  # pragma: no cover
                "Unable to compute differences between {}{} - {}{}\n{}\n"
                "--------\n{}".format(
                    skl_pred.shape, " (sparse)" if skl_sparse else "",
                    ort_pred.shape, " (sparse)" if ort_pred else "",
                    skl_pred, ort_pred))
        return rel_diff



"""
@file
@brief Modified converter from
`XGBoost.py <https://github.com/onnx/onnxmltools/blob/master/onnxmltools/convert/
xgboost/operator_converters/XGBoost.py>`_.
"""
import json
from pprint import pformat
import numpy
from xgboost import XGBClassifier
from skl2onnx.common.data_types import guess_numpy_type  # pylint: disable=C0411


class XGBConverter:
    "common methods for converters"

    @staticmethod
    def get_xgb_params(xgb_node):
        """
        Retrieves parameters of a model.
        """
        pars = xgb_node.get_xgb_params()
        # xgboost >= 1.0
        if 'n_estimators' not in pars:
            pars['n_estimators'] = xgb_node.n_estimators
        return pars

    @staticmethod
    def validate(xgb_node):
        "validates the model"
        params = XGBConverter.get_xgb_params(xgb_node)
        try:
            if "objective" not in params:
                raise AttributeError('ojective')
        except AttributeError as e:
            raise RuntimeError('Missing attribute in XGBoost model.') from e

    @staticmethod
    def common_members(xgb_node, inputs):
        "common to regresssor and classifier"
        params = XGBConverter.get_xgb_params(xgb_node)
        objective = params["objective"]
        base_score = params["base_score"]
        booster = xgb_node.get_booster()
        # The json format was available in October 2017.
        # XGBoost 0.7 was the first version released with it.
        js_tree_list = booster.get_dump(with_stats=True, dump_format='json')
        js_trees = [json.loads(s) for s in js_tree_list]
        return objective, base_score, js_trees

    @staticmethod
    def _get_default_tree_attribute_pairs(is_classifier):
        attrs = {}
        for k in {'nodes_treeids', 'nodes_nodeids',
                  'nodes_featureids', 'nodes_modes', 'nodes_values',
                  'nodes_truenodeids', 'nodes_falsenodeids', 'nodes_missing_value_tracks_true'}:
            attrs[k] = []
        if is_classifier:
            for k in {'class_treeids', 'class_nodeids', 'class_ids', 'class_weights'}:
                attrs[k] = []
        else:
            for k in {'target_treeids', 'target_nodeids', 'target_ids', 'target_weights'}:
                attrs[k] = []
        return attrs

    @staticmethod
    def _add_node(attr_pairs, is_classifier, tree_id, tree_weight, node_id,
                  feature_id, mode, value, true_child_id, false_child_id, weights, weight_id_bias,
                  missing, hitrate):
        if isinstance(feature_id, str):
            # Something like f0, f1...
            if feature_id[0] == "f":
                try:
                    feature_id = int(feature_id[1:])
                except ValueError as e:  # pragma: no cover
                    raise RuntimeError(
                        "Unable to interpret '{0}'".format(feature_id)) from e
            else:  # pragma: no cover
                try:
                    feature_id = int(feature_id)
                except ValueError:
                    raise RuntimeError(
                        "Unable to interpret '{0}'".format(feature_id)) from e

        # Split condition for sklearn
        # * if X_ptr[X_sample_stride * i + X_fx_stride * node.feature] <= node.threshold:
        # * https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/tree/_tree.pyx#L946
        # Split condition for xgboost
        # * if (fvalue < split_value)
        # * https://github.com/dmlc/xgboost/blob/master/include/xgboost/tree_model.h#L804

        attr_pairs['nodes_treeids'].append(tree_id)
        attr_pairs['nodes_nodeids'].append(node_id)
        attr_pairs['nodes_featureids'].append(feature_id)
        attr_pairs['nodes_modes'].append(mode)
        attr_pairs['nodes_values'].append(float(value))
        attr_pairs['nodes_truenodeids'].append(true_child_id)
        attr_pairs['nodes_falsenodeids'].append(false_child_id)
        attr_pairs['nodes_missing_value_tracks_true'].append(missing)
        if 'nodes_hitrates' in attr_pairs:
            attr_pairs['nodes_hitrates'].append(hitrate)  # pragma: no cover
        if mode == 'LEAF':
            if is_classifier:
                for i, w in enumerate(weights):
                    attr_pairs['class_treeids'].append(tree_id)
                    attr_pairs['class_nodeids'].append(node_id)
                    attr_pairs['class_ids'].append(i + weight_id_bias)
                    attr_pairs['class_weights'].append(float(tree_weight * w))
            else:
                for i, w in enumerate(weights):
                    attr_pairs['target_treeids'].append(tree_id)
                    attr_pairs['target_nodeids'].append(node_id)
                    attr_pairs['target_ids'].append(i + weight_id_bias)
                    attr_pairs['target_weights'].append(float(tree_weight * w))

    @staticmethod
    def _fill_node_attributes(treeid, tree_weight, jsnode, attr_pairs, is_classifier, remap):
        if 'children' in jsnode:
            XGBConverter._add_node(attr_pairs=attr_pairs, is_classifier=is_classifier,
                                   tree_id=treeid, tree_weight=tree_weight,
                                   value=jsnode['split_condition'], node_id=remap[jsnode['nodeid']],
                                   feature_id=jsnode['split'],
                                   mode='BRANCH_LT',  # 'BRANCH_LEQ' --> is for sklearn
                                   # ['children'][0]['nodeid'],
                                   true_child_id=remap[jsnode['yes']],
                                   # ['children'][1]['nodeid'],
                                   false_child_id=remap[jsnode['no']],
                                   weights=None, weight_id_bias=None,
                                   # ['children'][0]['nodeid'],
                                   missing=jsnode.get(
                                       'missing', -1) == jsnode['yes'],
                                   hitrate=jsnode.get('cover', 0))

            for ch in jsnode['children']:
                if 'children' in ch or 'leaf' in ch:
                    XGBConverter._fill_node_attributes(
                        treeid, tree_weight, ch, attr_pairs, is_classifier, remap)
                else:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to convert this node {0}".format(ch))

        else:
            weights = [jsnode['leaf']]
            weights_id_bias = 0
            XGBConverter._add_node(attr_pairs=attr_pairs, is_classifier=is_classifier,
                                   tree_id=treeid, tree_weight=tree_weight,
                                   value=0., node_id=remap[jsnode['nodeid']],
                                   feature_id=0, mode='LEAF',
                                   true_child_id=0, false_child_id=0,
                                   weights=weights, weight_id_bias=weights_id_bias,
                                   missing=False, hitrate=jsnode.get('cover', 0))

    @staticmethod
    def _remap_nodeid(jsnode, remap=None):
        if remap is None:
            remap = {}
        nid = jsnode['nodeid']
        remap[nid] = len(remap)
        if 'children' in jsnode:
            for ch in jsnode['children']:
                XGBConverter._remap_nodeid(ch, remap)
        return remap

    @staticmethod
    def fill_tree_attributes(js_xgb_node, attr_pairs, tree_weights, is_classifier):
        "fills tree attributes"
        if not isinstance(js_xgb_node, list):
            raise TypeError(  # pragma: no cover
                "js_xgb_node must be a list")
        for treeid, (jstree, w) in enumerate(zip(js_xgb_node, tree_weights)):
            remap = XGBConverter._remap_nodeid(jstree)
            XGBConverter._fill_node_attributes(
                treeid, w, jstree, attr_pairs, is_classifier, remap)


class XGBRegressorConverter(XGBConverter):
    "converter class"

    @staticmethod
    def validate(xgb_node):
        return XGBConverter.validate(xgb_node)

    @staticmethod
    def _get_default_tree_attribute_pairs():  # pylint: disable=W0221
        attrs = XGBConverter._get_default_tree_attribute_pairs(False)
        attrs['post_transform'] = 'NONE'
        attrs['n_targets'] = 1
        return attrs

    @staticmethod
    def convert(scope, operator, container):
        "converter method"
        dtype = guess_numpy_type(operator.inputs[0].type)
        if dtype != numpy.float64:
            dtype = numpy.float32
        xgb_node = operator.raw_operator
        inputs = operator.inputs
        objective, base_score, js_trees = XGBConverter.common_members(
            xgb_node, inputs)

        if objective in ["reg:gamma", "reg:tweedie"]:
            raise RuntimeError(  # pragma: no cover
                "Objective '{}' not supported.".format(objective))

        booster = xgb_node.get_booster()
        if booster is None:
            raise RuntimeError(  # pragma: no cover
                "The model was probably not trained.")

        best_ntree_limit = getattr(booster, 'best_ntree_limit', len(js_trees))
        if best_ntree_limit < len(js_trees):
            js_trees = js_trees[:best_ntree_limit]

        attr_pairs = XGBRegressorConverter._get_default_tree_attribute_pairs()
        attr_pairs['base_values'] = [base_score]
        XGBConverter.fill_tree_attributes(
            js_trees, attr_pairs, [1 for _ in js_trees], False)

        # add nodes
        if dtype == numpy.float64:
            container.add_node('TreeEnsembleRegressorDouble', operator.input_full_names,
                               operator.output_full_names,
                               name=scope.get_unique_operator_name(
                                   'TreeEnsembleRegressorDouble'),
                               op_domain='mlprodict', **attr_pairs)
        else:
            container.add_node('TreeEnsembleRegressor', operator.input_full_names,
                               operator.output_full_names,
                               name=scope.get_unique_operator_name(
                                   'TreeEnsembleRegressor'),
                               op_domain='ai.onnx.ml', **attr_pairs)


class XGBClassifierConverter(XGBConverter):
    "converter for XGBClassifier"

    @staticmethod
    def validate(xgb_node):
        return XGBConverter.validate(xgb_node)

    @staticmethod
    def _get_default_tree_attribute_pairs():  # pylint: disable=W0221
        attrs = XGBConverter._get_default_tree_attribute_pairs(True)
        # attrs['nodes_hitrates'] = []
        return attrs

    @staticmethod
    def convert(scope, operator, container):
        "convert method"
        dtype = guess_numpy_type(operator.inputs[0].type)
        if dtype != numpy.float64:
            dtype = numpy.float32
        xgb_node = operator.raw_operator
        inputs = operator.inputs

        objective, base_score, js_trees = XGBConverter.common_members(
            xgb_node, inputs)
        if base_score is None:
            raise RuntimeError(  # pragma: no cover
                "base_score cannot be None")
        params = XGBConverter.get_xgb_params(xgb_node)

        attr_pairs = XGBClassifierConverter._get_default_tree_attribute_pairs()
        XGBConverter.fill_tree_attributes(
            js_trees, attr_pairs, [1 for _ in js_trees], True)

        ncl = (max(attr_pairs['class_treeids']) + 1) // params['n_estimators']

        bst = xgb_node.get_booster()
        best_ntree_limit = getattr(
            bst, 'best_ntree_limit', len(js_trees)) * ncl
        if best_ntree_limit < len(js_trees):
            js_trees = js_trees[:best_ntree_limit]
            attr_pairs = XGBClassifierConverter._get_default_tree_attribute_pairs()
            XGBConverter.fill_tree_attributes(
                js_trees, attr_pairs, [1 for _ in js_trees], True)

        if len(attr_pairs['class_treeids']) == 0:
            raise RuntimeError(  # pragma: no cover
                "XGBoost model is empty.")
        if 'n_estimators' not in params:
            raise RuntimeError(  # pragma: no cover
                "Parameters not found, existing:\n{}".format(
                    pformat(params)))
        if ncl <= 1:
            ncl = 2
            # See https://github.com/dmlc/xgboost/blob/master/src/common/math.h#L23.
            attr_pairs['post_transform'] = "LOGISTIC"
            attr_pairs['class_ids'] = [0 for v in attr_pairs['class_treeids']]
        else:
            # See https://github.com/dmlc/xgboost/blob/master/src/common/math.h#L35.
            attr_pairs['post_transform'] = "SOFTMAX"
            # attr_pairs['base_values'] = [base_score for n in range(ncl)]
            attr_pairs['class_ids'] = [v % ncl
                                       for v in attr_pairs['class_treeids']]

        classes = xgb_node.classes_
        if (numpy.issubdtype(classes.dtype, numpy.floating) or
                numpy.issubdtype(classes.dtype, numpy.signedinteger)):
            attr_pairs['classlabels_int64s'] = classes.astype('int')
        else:
            classes = numpy.array([s.encode('utf-8') for s in classes])
            attr_pairs['classlabels_strings'] = classes

        if dtype == numpy.float64:
            op_name = "TreeEnsembleClassifierDouble"
        else:
            op_name = "TreeEnsembleClassifier"

        # add nodes
        if objective == "binary:logistic":
            ncl = 2
            container.add_node(op_name, operator.input_full_names,
                               operator.output_full_names,
                               name=scope.get_unique_operator_name(
                                   op_name),
                               op_domain='ai.onnx.ml', **attr_pairs)
        elif objective == "multi:softprob":
            ncl = len(js_trees) // params['n_estimators']
            container.add_node(op_name, operator.input_full_names,
                               operator.output_full_names,
                               name=scope.get_unique_operator_name(
                                   op_name),
                               op_domain='ai.onnx.ml', **attr_pairs)
        elif objective == "reg:logistic":
            ncl = len(js_trees) // params['n_estimators']
            if ncl == 1:
                ncl = 2
            container.add_node(op_name, operator.input_full_names,
                               operator.output_full_names,
                               name=scope.get_unique_operator_name(
                                   op_name),
                               op_domain='ai.onnx.ml', **attr_pairs)
        else:
            raise RuntimeError(  # pragma: no cover
                "Unexpected objective: {0}".format(objective))


def convert_xgboost(scope, operator, container):
    """
    This converters reuses the code from
    `XGBoost.py <https://github.com/onnx/onnxmltools/blob/master/onnxmltools/convert/
    xgboost/operator_converters/XGBoost.py>`_ and makes
    some modifications. It implements converters
    for models in :epkg:`xgboost`.
    """
    xgb_node = operator.raw_operator
    if isinstance(xgb_node, XGBClassifier):
        cls = XGBClassifierConverter
    else:
        cls = XGBRegressorConverter
    cls.convert(scope, operator, container)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRunBinaryNum
from ._op_numpy_helper import numpy_dot_inplace


class MatMul(OpRunBinaryNum):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryNum.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, a, b):  # pylint: disable=W0221
        return (numpy_dot_inplace(self.inplaces, a, b), )

    def to_python(self, inputs):
        return "import numpy", "return %s @ %s" % tuple(inputs)



"""
@file
@brief Runtime operator.
"""
import numpy
from onnx import TensorProto


def _get_typed_class_attribute(self, k, atts):
    """
    Converts an attribute into a C++ value.
    """
    ty = atts[k]
    if isinstance(ty, numpy.ndarray):
        v = getattr(self, k)
        return v if v.dtype == ty.dtype else v.astype(ty.dtype)
    if isinstance(ty, bytes):
        return getattr(self, k).decode()
    if isinstance(ty, list):
        return [_.decode() for _ in getattr(self, k)]
    if isinstance(ty, int):
        return getattr(self, k)
    raise NotImplementedError(  # pragma: no cover
        "Unable to convert '{}' ({}).".format(
            k, getattr(self, k)))


def proto2dtype(proto_type):
    """
    Converts a proto type into a :epkg:`numpy` type.

    :param proto_type: example ``onnx.TensorProto.FLOAT``
    :return: :epkg:`numpy` dtype
    """
    if proto_type == TensorProto.FLOAT:  # pylint: disable=E1101
        return numpy.float32
    if proto_type == TensorProto.BOOL:  # pylint: disable=E1101
        return numpy.bool
    if proto_type == TensorProto.DOUBLE:  # pylint: disable=E1101
        return numpy.float64
    if proto_type == TensorProto.STRING:  # pylint: disable=E1101
        return numpy.str
    if proto_type == TensorProto.INT64:  # pylint: disable=E1101
        return numpy.int64
    if proto_type == TensorProto.INT32:  # pylint: disable=E1101
        return numpy.int32
    if proto_type == TensorProto.FLOAT16:  # pylint: disable=E1101
        return numpy.float16
    raise ValueError(
        "Unable to convert proto_type {} to numpy type.".format(
            proto_type))


def dtype_name(dtype):
    """
    Returns the name of a numpy dtype.
    """
    if dtype == numpy.float32:
        return "float32"
    if dtype == numpy.float64:
        return "float64"
    if dtype == numpy.float16:
        return "float16"
    if dtype == numpy.int32:
        return "int32"
    if dtype == numpy.int64:
        return "int64"
    if dtype == numpy.str:
        return "str"
    if dtype == numpy.bool:
        return "bool"
    raise ValueError(
        "Unexpected dtype {}.".format(dtype))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Where(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, condition, x, y):  # pylint: disable=W0221
        if x.dtype != y.dtype:
            raise RuntimeError(  # pragma: no cover
                "x and y should share the same dtype {} != {}".format(
                    x.dtype, y.dtype))
        if x.shape != y.shape:
            raise RuntimeError(  # pragma: no cover
                "x and y should share the same shape {} != {}".format(
                    x.shape, y.shape))
        return (numpy.where(condition, x, y).astype(x.dtype), )

    def _infer_shapes(self, condition, x, y):  # pylint: disable=W0221
        """
        Returns the same shape by default.
        """
        return (x, )



# -*- coding: utf-8 -*-
"""
@file
@brief List of converters from scikit-learn model.
"""
import numpy
from .g_sklearn_type_helpers import check_type
from .grammar.gactions import MLActionVar, MLActionCst, MLActionIfElse, MLActionReturn
from .grammar.gactions_tensor import MLActionTensorTake
from .grammar.gactions_num import MLActionTestInf, MLActionTestEqual
from .grammar.gmlactions import MLModel


def sklearn_decision_tree_regressor(model, input_names=None, output_names=None, **kwargs):
    """
    Converts a `DecisionTreeRegressor
    <http://scikit-learn.org/stable/modules/generated/
    sklearn.tree.DecisionTreeRegressor.html>`_
    model into a *grammar* model (semantic graph representation).

    @param      model           scikit-learn model
    @param      input_names     name of the input features
    @param      output_names    name of the output predictions
    @param      kwargs          addition parameter (*with_loop*)
    @return                     graph model

    If *input* is None or *output* is None, default values
    will be given to the outputs
    ``['Prediction', 'Score']`` for the outputs.
    If *input_names* is None, it wil be ``'Features'``.

    Additional parameters:
    - *with_loop*: False by default, *True* not implemented.

    .. note::

        The code to compute on output is
        `here <https://github.com/scikit-learn/scikit-learn/blob/
        ef5cb84a805efbe4bb06516670a9b8c690992bd7/sklearn/tree/_tree.pyx#L806>`_:

        ::

            for i in range(n_samples):
                node = self.nodes
                # While node not a leaf
                while node.left_child != _TREE_LEAF:
                    # ... and node.right_child != _TREE_LEAF:
                    if X_ptr[X_sample_stride * i +
                             X_fx_stride * node.feature] <= node.threshold:
                        node = &self.nodes[node.left_child]
                    else:
                        node = &self.nodes[node.right_child]

                out_ptr[i] = <SIZE_t>(node - self.nodes)  # node offset

    TODO: improve C code (all leaves are computed and this is unnecessary).
    TODO: create a function tree and an intermediate node and use it here.
    """
    if kwargs.get('with_loop', False):
        raise NotImplementedError(  # pragma: no cover
            "Loop version is not implemented.")
    if output_names is None:
        output_names = ['Prediction', 'Score']
    if input_names is None:
        input_names = 'Features'

    from sklearn.tree import DecisionTreeRegressor
    check_type(model, DecisionTreeRegressor)

    # We convert the tree into arrays.
    # run help(model.tree_).
    lthres = MLActionCst(model.tree_.threshold.ravel().astype(
        numpy.float32), comment="threshold")
    lleft = MLActionCst(model.tree_.children_left.ravel().astype(
        numpy.int32), comment="left")
    lright = MLActionCst(model.tree_.children_right.ravel().astype(
        numpy.int32), comment="right")
    lfeat = MLActionCst(model.tree_.feature.ravel().astype(
        numpy.int32), comment="indfeat")
    lvalue = MLActionCst(model.tree_.value.ravel().astype(
        numpy.float32), comment="value")

    ex = numpy.zeros(model.n_features_, numpy.float32)
    lvar = MLActionVar(ex, input_names)

    lind = MLActionCst(numpy.int32(0), comment="lind")
    th = MLActionTensorTake(lthres, lind)
    m1 = MLActionCst(numpy.int32(-1), comment="m1")

    max_depth = model.tree_.max_depth
    cont = None
    new_lind = None
    for i in range(0, max_depth):
        # Leave ?
        if new_lind is not None:
            lind = new_lind

        le = MLActionTensorTake(lleft, lind)
        lr = MLActionTensorTake(lright, lind)

        di = MLActionTensorTake(lfeat, lind)
        df = MLActionTensorTake(lfeat, di)
        xx = MLActionTensorTake(lvar, df)
        te = MLActionTestInf(xx, th)

        new_lind = MLActionIfElse(te, le, lr, comment="lind{0}".format(i))
        le = MLActionTensorTake(lleft, new_lind)
        th = MLActionTensorTake(lthres, new_lind)

        eq = MLActionTestEqual(m1, le)
        va = MLActionTensorTake(lvalue, new_lind)
        cont = MLActionIfElse(eq, va, th, comment="cont{0}".format(i))

    ret = MLActionReturn(cont)
    return MLModel(ret, output_names, name=DecisionTreeRegressor.__name__)



"""
@file
@brief Functions to help guessing the final graph structure.
"""
import numpy
from onnxconverter_common.data_types import Float16TensorType
from skl2onnx.common.data_types import (
    DataType,
    FloatTensorType, SequenceType, DictionaryType,
    Int64Type, Int64TensorType, BooleanTensorType,
    Int32TensorType, DoubleTensorType, FloatType,
    StringTensorType)
from skl2onnx.common.data_types import _guess_type_proto
from skl2onnx.algebra.type_helper import _guess_type as skl2onnx__guess_type
from skl2onnx.proto import TensorProto


def _guess_type(var):
    if isinstance(var, dict) and 'value' in var:
        return skl2onnx__guess_type(var['value'])
    return skl2onnx__guess_type(var)


def get_defined_inputs(input_names, variables=None, dtype=None):
    """
    Retrieves defined inputs in already declared variables
    bsed on their names.

    @param      input_names     input names
    @param      variables       registered variables created
                                by previous operators
    @param      dtype           float computational type
    @return                     typed inputs
                                as ``tuple(name, type)``
    """
    def guess_type_variable(name):
        if variables is None:
            return DoubleTensorType() if dtype == numpy.float64 else FloatTensorType()
        elif name in variables:
            ty = variables[name]
            if isinstance(ty, DataType):
                shape = ty.shape
                if 0 in shape:
                    raise RuntimeError(  # pragma: no cover
                        "Shape cannot be empty: name='{}', var={}".format(
                            name, ty))
                return variables[name]
            if isinstance(ty, dict) and 'value' in ty:
                # constant
                arr = ty['value']
                try:
                    return _guess_type(arr)
                except RuntimeError as e:  # pragma: no cover
                    raise RuntimeError(
                        "Unable to guess type of variable '{}' - {}."
                        "".format(name, arr)) from e
            raise NotImplementedError(  # pragma: no cover
                "Unable to guess type for '{}' form '{}'.".format(
                    name, variables[name]))
        else:
            # Inputs. Let's assume it is a vector of floats.
            return DoubleTensorType() if dtype == numpy.float64 else FloatTensorType()

    inputs = [(name, guess_type_variable(name)) for name in input_names]
    return inputs


def get_defined_outputs(outputs, onnx_node, typed_inputs=None, variables=None, dtype=None):
    """
    Gets types of predefined outputs when they cannot be inferred.
    Some part of it should be automated based
    on type constraints.

    @param      outputs         requested outputs
    @param      onnx_node       :epkg:`ONNX` node definition
    @param      typed_inputs    known typed inputs of the node
                                as ``tuple(name, type)``
    @param      variables       registered variables created
                                by previous operators
    @param      dtype           float computational type
    @return                     typed outputs
                                as ``tuple(name, type)``
    """
    ft = DoubleTensorType if dtype == numpy.float64 else FloatTensorType

    # ZipMap
    if onnx_node.op_type == "ZipMap":
        otype = SequenceType(DictionaryType(
            Int64Type(), ft()))
        outputs = [(name, otype) for name in outputs]
    # ArgMin, ArgMax, Shape
    elif onnx_node.op_type in ("ArgMin", "ArgMax", 'Shape') and len(outputs) == 1:
        outputs = [(outputs[0], Int64TensorType())]
    # Greater, Less, Equal
    elif onnx_node.op_type in ("Greater", "Less", 'Equal') and len(outputs) == 1:
        outputs = [(outputs[0], BooleanTensorType())]
    # TopK
    elif onnx_node.op_type == "TopK" and len(outputs) == 2:
        if len(typed_inputs) != 2:
            raise RuntimeError(  # pragma: no cover
                "Wrong typed_inputs, got {}.".format(typed_inputs))
        outputs = [(outputs[0], typed_inputs[0][1]),
                   (outputs[1], Int64TensorType())]
    # Cast
    elif onnx_node.op_type == "Cast" and len(outputs) == 1:
        ttyp = _guess_type_proto(onnx_node.attribute[0].i, dims=None)
        outputs = [(outputs[0], ttyp)]
    # ArrayFeatureExtractor
    elif onnx_node.op_type == "ArrayFeatureExtractor":
        if len(typed_inputs) != 2:
            raise RuntimeError(  # pragma: no cover
                "Wrong typed_inputs, got {}.".format(typed_inputs))
        outputs = [(outputs[0], typed_inputs[0][1])]
    elif 'Classifier' in onnx_node.op_type:
        # Good chance that's a classifier.
        outputs = [(outputs[0], Int64TensorType()),
                   (outputs[1], ft())]
    # Reshape
    elif onnx_node.op_type in ('Reshape', 'Transpose'):
        outputs = [(outputs[0], typed_inputs[0][1].__class__())]
    # Scan
    elif onnx_node.op_type == 'Scan':
        if len(outputs) != len(typed_inputs):
            raise RuntimeError(  # pragma: no cover
                "Dimension mismatch, operator Scan should have "
                "the same number of inputs and outputs {} != {}"
                ".".format(len(outputs), len(typed_inputs)))
        outputs = [(o, t[1].__class__())
                   for o, t in zip(outputs, typed_inputs)]
    # ConstantOfShape
    elif onnx_node.op_type == "ConstantOfShape":
        outputs = [(outputs[0], ft())]

    # Default case
    # Assuming the only output is the same as the only input.
    elif len(typed_inputs) == 1 and len(outputs) == 1:
        outputs = [(outputs[0], typed_inputs[0][1])]
    # Default
    else:
        outputs = [(name, ft()) for name in outputs]
    return outputs


def proto2vars(values):
    """
    Converts proto values to Variables.
    """
    def ptype2vttype(it, shape):
        if it == TensorProto.FLOAT:  # pylint: disable=E1101
            return FloatTensorType(shape)
        if it == TensorProto.DOUBLE:  # pylint: disable=E1101
            return DoubleTensorType(shape)
        if it == TensorProto.INT64:  # pylint: disable=E1101
            return Int64TensorType(shape)
        if it == TensorProto.INT32:  # pylint: disable=E1101
            return Int32TensorType(shape)
        if it == TensorProto.BOOL:  # pylint: disable=E1101
            return BooleanTensorType(shape)
        if it == TensorProto.STRING:  # pylint: disable=E1101
            return StringTensorType(shape)
        if it == TensorProto.FLOAT16:  # pylint: disable=E1101
            return Float16TensorType(shape)
        raise NotImplementedError(  # pragma: no cover
            "Unrecognized proto type {} with shape {}".format(it, shape))

    def ptype2vtype(it):
        if it == TensorProto.FLOAT:  # pylint: disable=E1101
            return FloatType()
        if it == TensorProto.INT64:  # pylint: disable=E1101
            return Int64Type()
        raise NotImplementedError(  # pragma: no cover
            "Unrecognized proto type {}".format(it))

    res = []
    for v_ in values:
        v = v_
        name = v.name if hasattr(v, 'name') else None
        if hasattr(v, 'type') and str(v.type) != '':
            t = v.type
            v = proto2vars([t])[0][1]
        elif hasattr(v, 'sequence_type') and str(v.sequence_type) != '':
            subtype = proto2vars([v.sequence_type.elem_type])[0][1]
            v = SequenceType(subtype)
        elif hasattr(v, 'tensor_type') and str(v.tensor_type) != '':
            tt = v.tensor_type
            el = tt.elem_type
            shape = tt.shape
            dim = shape.dim
            if len(dim) == 0:
                shape = []
            else:
                shape = [dim[i].dim_value for i in range(len(dim))]
            v = ptype2vttype(el, shape)
        elif hasattr(v, 'map_type') and str(v.map_type) != '':
            mt = v.map_type
            keyt = ptype2vtype(mt.key_type)
            valt = proto2vars([mt.value_type])[0][1]
            v = DictionaryType(keyt, valt)
        else:
            raise RuntimeError(  # pragma: no cover
                "Unable to build a variable from {}.".format(v))
        if v.shape is not None and 0 in v.shape:
            # Replaces 0 by None
            new_shape = tuple(None if d == 0 else d for d in v.shape)
            if new_shape in ((None, ), None):
                v = v.__class__()
            else:
                v = v.__class__(new_shape)
        if v.shape is not None and 0 in v.shape:
            raise RuntimeError(  # pragma: no cover
                "Shape cannot be empty: '{}': {}.".format(
                    name, v_))
        res.append((name, v))
    return res



"""
@file
@brief A couple of tools related to filenames.
"""
import os


def extract_information_from_filename(name):
    """
    Returns a dictionary with information extracted
    from a filename.
    An example is better:

    .. runpython::
        :showcode:

        from mlprodict.tools.filename_helper import extract_information_from_filename

        candidates = [
            'bench_DecisionTreeClassifier_default_b_cl_1_4_12_float_.py',
            'bench_DecisionTreeClassifier_default_b_cl_64_10_20_12_double_.py',
            'bench_DecisionTreeClassifier_default_b_cl_64_100_4_12_float_.py',
            'bench_AdaBoostClassifier_default_b_cl_1000_50_12_float__fct.svg',
            'bench_AdaBoostClassifier_default_m_cl_1_4_12_float__line.svg',
            'bench_LogisticRegression_liblinear_b_cl_solverliblinear_1_4_12_float_nozipmap_fct.svg',
        ]

        for name in candidates:
            d = extract_information_from_filename(name)
            print(d)
    """
    spl = os.path.splitext(os.path.split(name)[-1])[0].split('_')
    res = {}
    for v in spl:
        if v == "bench":
            continue
        if not v:
            continue
        if "A" <= v[0] <= "Z":
            res['model'] = v
            continue
        try:
            i = int(v)
        except ValueError:
            i = None

        if i is not None:
            if i == 64:
                res['double'] = True
                continue
            if 'N' not in res:
                res['N'] = i
                continue
            if 'nf' not in res:
                res['nf'] = i
                continue
            if 'opset' not in res:
                res['opset'] = i
                continue
            raise ValueError(  # pragma: no cover
                "Unable to parse '{}'.".format(name))

        if 'scenario' not in res:
            res['scenario'] = v
            continue
        if 'N' in res:
            if v in ('fct', 'line'):
                res['profile'] = v
                continue
            res['opt'] = res.get('opt', '') + '_' + v
            continue
        if len(v) <= 4:
            res['problem'] = res.get('problem', '') + '_' + v
        else:
            res['opt'] = res.get('opt', '') + '_' + v

    for k in res:
        if isinstance(res[k], str):
            res[k] = res[k].strip('_')

    rep = {
        'LinReg': 'LinearRegression',
        'LinRegressor': 'LinearRegression',
        'LogReg': 'LogisticRegression',
        'HGB': 'HistGradientBoosting',
    }

    if 'model' in res:
        if res['model'].endswith('Clas'):
            res['model'] += "sifier"
        elif res['model'].endswith('Reg'):
            res['model'] += "ressor"
        if res['model'].startswith('HGB'):
            res['model'] = "HistGradientBoosting" + \
                res['model'][3:]  # pragma: no cover
        res['model'] = rep.get(res['model'], res['model'])
    return res


def make_readable_title(infos):
    """
    Creates a readable title based on the test information.
    """
    sp = [infos['model']]
    if 'problem' in infos:
        sp.append('[{}]'.format(infos['problem']))
    if 'scenario' in infos:
        sp.append('[{}]'.format(infos['scenario']))
    if 'N' in infos:
        sp.append('N={}'.format(infos['N']))
    if 'nf' in infos:
        sp.append('nf={}'.format(infos['nf']))
    if 'opset' in infos:
        sp.append('ops={}'.format(infos['opset']))
    if 'double' in infos:
        if infos['double']:
            sp.append('x64')
    if 'opt' in infos:
        sp.append('[{}]'.format(infos['opt']))
    if 'profile' in infos:
        sp.append('by {}'.format(infos['profile']))
    return " ".join(sp)



"""
@file
@brief Utilies to test script from :epkg:`scikit-learn` documentation.
"""
import os
from io import StringIO
from contextlib import redirect_stdout, redirect_stderr
import pprint
import numpy
from sklearn.base import BaseEstimator
from .verify_code import verify_code


class MissingVariableError(RuntimeError):
    """
    Raised when a variable is missing.
    """
    pass


def _clean_script(content):
    """
    Comments out all lines containing ``.show()``.
    """
    new_lines = []
    for line in content.split('\n'):
        if '.show()' in line or 'sys.exit' in line:
            new_lines.append("# " + line)
        else:
            new_lines.append(line)
    return "\n".join(new_lines)


def _enumerate_fit_info(fits):
    """
    Extracts the name of the fitted models and the data
    used to train it.
    """
    for fit in fits:
        chs = fit['children']
        if len(chs) < 2:
            # unable to extract the needed information
            continue  # pragma: no cover
        model = chs[0]['str']
        if model.endswith('.fit'):
            model = model[:-4]
        args = [ch['str'] for ch in chs[1:]]
        yield model, args


def _try_onnx(loc, model_name, args_name, **options):
    """
    Tries onnx conversion.

    @param  loc         available variables
    @param  model_name  model name among these variables
    @param  args_name   arguments name among these variables
    @param  options     additional options for the conversion
    @return             onnx model
    """
    from ..onnx_conv import to_onnx
    if model_name not in loc:
        raise MissingVariableError("Unable to find model '{}' in {}".format(
            model_name, ", ".join(sorted(loc))))
    if args_name[0] not in loc:
        raise MissingVariableError("Unable to find data '{}' in {}".format(
            args_name[0], ", ".join(sorted(loc))))
    model = loc[model_name]
    X = loc[args_name[0]]
    dtype = options.get('dtype', numpy.float32)
    Xt = X.astype(dtype)
    onx = to_onnx(model, Xt, **options)
    args = dict(onx=onx, model=model, X=Xt)
    return onx, args


def verify_script(file_or_name, try_onnx=True, existing_loc=None,
                  **options):
    """
    Checks that models fitted in an example from :epkg:`scikit-learn`
    documentation can be converted into :epkg:`ONNX`.

    @param      file_or_name        file or string
    @param      try_onnx            try the onnx conversion
    @param      existing_loc        existing local variables
    @param      options             conversion options
    @return                         list of converted models
    """
    if '\n' not in file_or_name and os.path.exists(file_or_name):
        filename = file_or_name
        with open(file_or_name, 'r', encoding='utf-8') as f:
            content = f.read()
    else:  # pragma: no cover
        content = file_or_name
        filename = "<string>"

    # comments out .show()
    content = _clean_script(content)

    # look for fit or predict expressions
    _, node = verify_code(content, exc=False)
    fits = node._fits
    models_args = list(_enumerate_fit_info(fits))

    # execution
    obj = compile(content, filename, 'exec')
    glo = globals().copy()
    loc = {}
    if existing_loc is not None:
        loc.update(existing_loc)
        glo.update(existing_loc)
    out = StringIO()
    err = StringIO()

    with redirect_stdout(out):
        with redirect_stderr(err):
            exec(obj, glo, loc)  # pylint: disable=W0122

    # filter out values
    cls = (BaseEstimator, numpy.ndarray)
    loc_fil = {k: v for k, v in loc.items() if isinstance(v, cls)}
    glo_fil = {k: v for k, v in glo.items() if k not in {'__builtins__'}}
    onx_info = []

    # onnx
    if try_onnx:
        if len(models_args) == 0:
            raise MissingVariableError(
                "No detected trained model in '{}'\n{}\n--LOCALS--\n{}".format(
                    filename, content, pprint.pformat(loc_fil)))
        for model_args in models_args:
            try:
                onx, args = _try_onnx(loc_fil, *model_args, **options)
            except MissingVariableError as e:
                raise MissingVariableError("Unable to find variable in '{}'\n{}".format(
                    filename, pprint.pformat(fits))) from e
            loc_fil[model_args[0] + "_onnx"] = onx
            onx_info.append(args)

    # final results
    return dict(locals=loc_fil, globals=glo_fil,
                stdout=out.getvalue(),
                stderr=err.getvalue(),
                onx_info=onx_info)
